
/**************************************************************************
 *
 * Filename: affy.h
 * 
 * Purpose: Top-level header file for libaffy.
 *
 * Creation date: 4 April, 2005
 *
 * Author: Steven Eschrich
 *
 *
 * Update History
 * --------------
 * 4/4/05:  Import from old libaffy (AMH)
 * 4/8/05:  DAT pixeldata now a 2D array of unsigned ints (AMH) 
 * 4/19/05: Addy AFFY_CELL and AFFY_PIXREGION types (AMH)
 *
 **************************************************************************/

#ifndef _AFFY_H_
#define _AFFY_H_

/* All system-level includes needed by libaffy are included here. */
#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>
#include <assert.h>
#include <float.h>
#include <math.h>
#include <time.h>
#include <dirent.h>

#include <affy_basetypes.h>
#include <utils_types.h>
#include "buffered_reader.h"

/* Types of locations on an affy chip. */
#define AFFY_UNDEFINED_LOCATION 0
#define AFFY_CONTROL_LOCATION   1
#define AFFY_QC_LOCATION        2
#define AFFY_NORMAL_LOCATION    3

/* Binary-format magic numbers. */
#define AFFY_DAT_FILEMAGIC 0xFC
#define AFFY_CDF_BINARYFILE_MAGIC 67
#define AFFY_CEL_BINARYFILE_MAGIC 64
#define AFFY_CEL_AGCC_MAGIC 59

/**************************************************************************/

/** High-level types. **/

/*
 * The CDF file contains most of the semantic information about the
 * affy chips. Therefore, there are many structures setup for
 * manipulating this information.
 *
 * First, CEL files and beyond are simply represented as a matrix of
 * data. The CDF provides all other necessary mapping (probes->probeset).
 *
 */

/* 
 * An AFFY_PIXREGION defines a subset or window into a grid of pixel
 * values.  **data is a 2D array of the dimensions given.
 */
typedef struct affy_pixregion_s
{
  affy_uint16 numrows;
  affy_uint16 numcols;
  unsigned int **data;          /* Must hold at least 16 bits; C89 says so. */
} AFFY_PIXREGION;

/*
 * A point somewhere on a cel or pixel map.
 */
typedef struct affy_point_s
{
  affy_int16 x;
  affy_int16 y;
} AFFY_POINT;

/*
 * An individual cell obtained from a CEL file.  A cell consists of some
 * number of pixels which can be accessed directly if the DAT file is
 * available.  
 */
typedef struct affy_cell_s
{
  double value;                 /* Intensity (mean pixel intensity).     */
  double stddev;                /* Std. deviation of pixel intensity.    */
  int numpixels;                /* Number of pixels composing this cell. */
  AFFY_PIXREGION *pixels;       /* The actual pixels making up the cell. */
} AFFY_CELL;

/*
 * A probe. The basic unit of information, it consists of both a 
 * PM (perfect match) and MM (mismatch) value.
 */
typedef struct affy_probe_s
{
  int index;                    /* Unique id for probe                  */
  AFFY_POINT mm;                /* Location of Mismatch value           */
  AFFY_POINT pm;                /* Location of Perfect match value      */
  struct affy_probeset_s *ps;   /* Pointer back to parent probe set     */
} AFFY_PROBE;

/* 
 * A ProbeSet consists of a set of individual probes. Each probe set
 * will eventually consist of a single expression value computed from
 * the various pm/mm probes.
 */
typedef struct affy_probeset_s
{
  int index;                    /* Unique id for probe set                */
  char *name;          /* A text description of probe set        */
  int numprobes;                /* Total number of probes                 */
  AFFY_PROBE *probe;           /* The probes themselves                  */
} AFFY_PROBESET;

/*
 * CDF file definitions. The CDF contains the meta-information about
 * a particular microarray chip, so there are many definitions in this
 * section. They link all of the various higher level details together.
 *
 * The CDF file consists of mappings from x,y coordinates into probes,
 * which are grouped by probe sets.
 */
typedef struct affy_cdffile_s
{
  char *array_type;							/* Name of file/chip type */
  int numrows;
  int numcols;
  int numprobes;                /* Count of total probes                   */
  int numprobesets;             /* Number of probe sets                    */
  int numqcunits;               /* Num quality control units               */

  affy_uint8 **cell_type;       /* What kind of cell (normal, qc, control) */
  AFFY_PROBE ***xy_ref;         /* row,col map to individual probe */
  AFFY_PROBESET *probeset;      /* An array of probe sets, made of probes  */
  AFFY_PROBE **probe;         	/* A linear array of probes                */
} AFFY_CDFFILE;

/*
 * Given the above definitions, a cel file can be a simple entity.
 */
typedef struct affy_celfile_s
{
  char *filename;
  int numrows;
  int numcols;
  int nummasks;
  int numoutliers;
  AFFY_CELL **data;
  affy_uint8 **mask;
  affy_uint8 **outlier;
} AFFY_CELFILE;

/*
 * DAT file definition; a binary representation of individual
 * pixel intensity values.  The raw data for the higher level files.
 */
typedef struct affy_datfile_s
{
  char *experiment_name;

/*  affy_uint16     pixels_per_line; */

/*  affy_uint16     numlines;        */

/*  affy_uint32     pixels_per_row;  */

/*  affy_uint32     numrows;         */
  affy_uint16 pixel_width;
  affy_uint16 pixel_height;
  affy_uint16 scanspeed;
  affy_float32 temperature;
  affy_float32 laser_power;
  char timestamp[19];           /* field width 18, plus NUL */
  affy_uint32 numpixels;
  affy_uint32 minpixel;
  affy_uint32 maxpixel;
  affy_uint32 numsamples_dc_offset;
  AFFY_POINT grid_ul;
  AFFY_POINT grid_ur;
  AFFY_POINT grid_ll;
  AFFY_POINT grid_lr;
  affy_uint16 cellmargin;
  char *scannerid;
  char *probe_array_type;
  affy_float64 meanpixel;
  affy_float64 std_dev_pixel;
  affy_float64 avg_dc_offset;
  affy_float64 std_dev_dc_offset;
  AFFY_PIXREGION pixels;
} AFFY_DATFILE;

/*
 * We need a higher abstraction than a celfile, to pull everything 
 * together and provide a convenient storage place for expressions, etc.
 *
 * This is a CHIP, or more generally (and usefully) a collection of one
 * or more CHIPS.
 *
 * A single chip consists of pointers to the meta-information and
 * actual information (CEL file). The summary information is stored
 * in a single array probe_set.
 */
typedef struct affy_chip_s
{
  char *filename;               /* Filename of chip                   */
  AFFY_CDFFILE *cdf;            /* The chip description file          */
  AFFY_CELFILE *cel;            /* For the CEL file information       */
  AFFY_DATFILE *dat;            /* Underlying pixel data (if present) */

  /* The true information - the probe sets */
  int numprobesets;
  double *probe_set;

  /* A convenience ptr: not globally used (RMA) */
  double *pm;
} AFFY_CHIP;

/*
 * And finally, a chipset is a group of the same type of chips. 
 */
typedef struct affy_chipset_s
{
  /* Some common parameters needed, from the CDF file. */
  int numchips;
  int numrows;
  int numcols;

  /* Each CHIPSET has a single description file (CDF) */
  AFFY_CDFFILE *cdf;

  /* 
   * These are the chips themselves. Note they are arrays of ptrs, 
   * so things can be copied more easily
   */
  AFFY_CHIP **chip;

  double *mean; /* storing the mean probe intensities */

} AFFY_CHIPSET;

/*
 * Probeset loading data structure (By Debashis Sahoo)
 */
typedef struct affy_ps_load_s {
  FILE *fp;
  int max_chips; /* Max number of chips */
  int numchips; /* Actual number of chips read successfully */
  AFFY_CHIPSET *chipset;
  struct hashtable * h_probesets;
  char *chiptype;
  buffered_reader **reader;
} AFFY_PSLOAD;

/**************************************************************************/

/** Function prototypes. **/

#ifdef __cplusplus
extern "C"
{
#endif

/* File I/O stuff. */
  AFFY_CDFFILE *affy_load_cdf_file(char *chip_type, char *dir);
  AFFY_CHIPSET *affy_load_data(char *dir, char **files);
  AFFY_CHIP *affy_load_chip(char *filename);
  AFFY_CELFILE *affy_load_cel_file(char *filename);
  AFFY_DATFILE *affy_load_dat_file(char *filename);
  void affy_free_cel_file(AFFY_CELFILE *cf);
  void affy_free_dat_file(AFFY_DATFILE *df);
  void affy_free_cdf_file(AFFY_CDFFILE *cdf);
  void affy_free_chip(AFFY_CHIP *ch);
  void affy_free_chipset(AFFY_CHIPSET *cs);
  char **affy_list_files(char *directory, char *extension);
  /* Added by DS */
  void save_pm_to_disk(AFFY_CHIPSET *c, int i, int max_chips, char *file);
  double * load_pm_from_disk(AFFY_CHIPSET *c, int curNum, int max_chips, char *file);
  int init_probeset_loading(char *file, AFFY_CHIPSET *c, AFFY_PSLOAD *ps);
  void close_probeset_loading(AFFY_PSLOAD *ps);
  void load_matrix_from_disk(AFFY_PSLOAD *psload, int ps, double **z);
  void save_mean_to_disk(double *mean, int numprobes, char *file);
  void load_mean_from_disk(double *mean, int *numprobes, char *file);
  char * load_filename_from_disk(AFFY_PSLOAD *psload, int curNum);
  struct hashtable * get_filenames(AFFY_PSLOAD *psload);
  struct hashtable * get_probesets(AFFY_CDFFILE *cdf);


/* Text handling functions. */
  char *affy_get_next_line(FILE * fp);
  void affy_unget_next_line(FILE * fp);
  void affy_reset_next_line(FILE * fp);
  void affy_skip_to_next_header(FILE * fp);

/* Binary handling functions. */
  int affy_read8(FILE * input, void *buf);
  int affy_read16(FILE * input, void *buf);
  int affy_read32(FILE * input, void *buf);
  int affy_read64(FILE * input, void *buf);
  int affy_readchars(FILE * input, char *buf, size_t numbytes);
	float affy_readfloat(FILE *input);
	int affy_readint(FILE *input);
	int affy_readbyte(FILE *input);
	int affy_readubyte(FILE *input);
	int affy_readshort(FILE *input);
	int affy_readushort(FILE *input);
	double affy_readdouble(FILE *input);
  int affy_read32_buffer(char * input, void *buf);
	/* Added by DS */
  int affy_write8(FILE * input, void *buf);
  int affy_write16(FILE * input, void *buf);
  int affy_write32(FILE * input, void *buf);
  int affy_write64(FILE * input, void *buf);
  int affy_writechars(FILE * input, char *buf, size_t numbytes);
	int affy_writefloat(FILE *input, float buf);
	int   affy_writeint(FILE *input, int buf);
	int affy_writedouble(FILE *input, double buf);
	
/* Utility functions */
  void affy_mean_normalization(AFFY_CHIPSET *d, double target_mean);
  char *affy_get_cdf_name(char *buf);
  char *affy_get_cdf_name_from_cel(char *filename);
  double **affy_matrix_from_cel(AFFY_CELFILE *cf);

  void affy_dump_dat_hdr(AFFY_DATFILE *df);
  void affy_write_expressions(AFFY_CHIPSET *c, char *filename);
  Boolean affy_ismasked(AFFY_CHIP *chip, int x, int y);
  Boolean affy_iscontrol(AFFY_CHIP *chip, int x, int y);
  Boolean affy_isqc(AFFY_CHIP *chip, int x, int y);
  Boolean affy_isundefined(AFFY_CHIP *chip, int x, int y);
  AFFY_PIXREGION *affy_pixels_from_cell(AFFY_CHIP *cp, int x, int y);
	AFFY_POINT affy_cell_to_pixel(AFFY_CHIP *cp, int x, int y);

	void affy_pixregion2tiff(AFFY_PIXREGION *p, char *filename);
	void affy_pixregion2text(AFFY_PIXREGION *p, char *filename);
	void affy_write_pixel_region(AFFY_PIXREGION *pr,char *filename);

/* Statistical functions (from util). */
  double affy_median_save(double *x, int length);
  double affy_median(double *x, int length);
  void affy_get_row_median(double **z, double *rdelta, int rows,
                           int columns);
  void affy_get_column_median(double **z, double *cdelta, int rows,
                              int columns);
  int affy_median_sort(const void *p1, const void *p2);
  int affy_qnorm_compare(const void *p1, const void *p2);
  void affy_rank_order(double *rank, double **x, int n);
  void affy_quantile_normalization(AFFY_CHIPSET *d);
  void affy_pnorm_both(double x, double *cum, double *ccum, int i_tail,
                       int log_p);
  double affy_pnorm5(double x, double mu, double sigma, int lower_tail,
                     int log_p);

#ifdef __cplusplus
};
#endif

#endif
