
/**************************************************************************
 *
 * Filename: binary_io.c
 * 
 * Purpose: Routines for reading discrete data types from a binary file.
 *          Endianness is accounted for and corrected.
 *
 * Creation date: 4 April, 2005
 *
 * Author: Andrew M. Hoerter
 *
 *
 * Update History
 * --------------
 * 4/4/05: Creation (AMH)
 *
 **************************************************************************/

#include <affy.h>
#include <endian_config.h>

/* Note that the endianness #define comes from the libutils package. */

/* Endian-swapping macros (x = (void *)). */
#ifdef LIBUTILS_BIG_ENDIAN
# define AFFY_SWAP16(x) *(affy_uint16 *)(x) =		               \
                         (((*(affy_uint16 *)(x)) & 0x00ff) << 8) |     \
			 (((*(affy_uint16 *)(x)) >> 8))

# define AFFY_SWAP32(x) *(affy_uint32 *)(x) =			       \
                         ((*(affy_uint32 *)(x) << 24) |	      	       \
                         ((*(affy_uint32 *)(x) << 8) & 0x00ff0000UL) | \
                         ((*(affy_uint32 *)(x) >> 8) & 0x0000ff00UL) | \
			 (*(affy_uint32 *)(x) >> 24))

# define AFFY_SWAPFL(x) AFFY_SWAP32(x)
# define AFFY_SWAPDBL(x) AFFY_SWAP64(x)

static void AFFY_SWAP64(void *buf)
{
  affy_uint8 *byte = (affy_uint8 *)buf;
  affy_uint8 tmp;

  tmp = byte[0];
  byte[0] = byte[7];
  byte[7] = tmp;

  tmp = byte[1];
  byte[1] = byte[6];
  byte[6] = tmp;

  tmp = byte[2];
  byte[2] = byte[5];
  byte[5] = tmp;

  tmp = byte[3];
  byte[3] = byte[4];
  byte[4] = tmp;
}

#else
# define AFFY_SWAP16(x)
# define AFFY_SWAP32(x)
# define AFFY_SWAPFL(x)
# define AFFY_SWAP64(x)
# define AFFY_SWAPDBL(x) AFFY_SWAP64(x)
#endif

/* 
 * affy_read8(): Read an unsigned byte off the file stream. 
 *
 * Inputs: *input is an open file input stream, *buf points to allocated
 *          storage large enough to hold an unsigned byte.
 * Outputs: 0 on success, -1 on failure (error or EOF).
 * Side effects: None.
 */
int affy_read8(FILE * input, void *buf)
{
  assert(input != NULL);
  assert(buf != NULL);

  if (fread(buf, 1, 1, input) < 1)
    return (-1);

  /* For bytes, that's it. */
  return (0);
}

/* 
 * affy_read16(): Read a 16-bit integer off the file stream. 
 *
 * Inputs: *input is an open file input stream, *buf points to allocated
 *          storage large enough to hold two bytes.
 * Outputs: 0 on success, -1 on failure (error or EOF).
 * Side effects: None.
 */
int affy_read16(FILE * input, void *buf)
{
  assert(input != NULL);
  assert(buf != NULL);

  if (fread(buf, 2, 1, input) < 1)
    return (-1);

  AFFY_SWAP16(buf);
  return (0);
}


/* 
 * affy_read32(): Read a 32-bit integer or float off the file stream. 
 *
 * Inputs: *input is an open file input stream, *buf points to allocated
 *          storage large enough to hold four bytes.
 * Outputs: 0 on success, -1 on failure (error or EOF).
 * Side effects: None.
 */
int affy_read32(FILE * input, void *buf)
{
  assert(input != NULL);
  assert(buf != NULL);

  if (fread(buf, 4, 1, input) < 1)
    return (-1);

  AFFY_SWAP32(buf);

  return (0);
}

/* 
 * affy_read32_buffer(): Read a 32-bit integer or float off the char stream. 
 *
 * Inputs: *input is a char input stream, *buf points to allocated
 *          storage large enough to hold four bytes.
 * Outputs: 0 on success, -1 on failure (error or EOF).
 * Side effects: None.
 */
int affy_read32_buffer(char * input, void *buf)
{
  assert(input != NULL);
  assert(buf != NULL);

  memcpy(buf, input, 4);

  AFFY_SWAP32(buf);

  return (0);
}

/* 
 * affy_read64(): Read a 64-bit integer or float off the file stream. 
 *
 * Inputs: *input is an open file input stream, *buf points to allocated
 *          storage large enough to hold eight bytes.
 * Outputs: 0 on success, -1 on failure (error or EOF).
 * Side effects: None.
 */
int affy_read64(FILE * input, void *buf)
{
  assert(input != NULL);
  assert(buf != NULL);

  if (fread(buf, 8, 1, input) < 1)
    return (-1);

  AFFY_SWAP64(buf);

  return (0);
}

/* 
 * affy_readchars(): Read a series of characters off the file stream. 
 *
 * Inputs: *input is an open file input stream, *buf points to allocated
 *          storage large enough to hold the desired number of bytes.
 *          At most numbytes-1 characters will be read, and a NUL byte
 *          will be appended.
 * Outputs: 0 on success, -1 on failure (error or EOF).
 * Side effects: None.
 *
 * Comments: Provided for ease of reading character strings.  Almost
 *           equivalent to fgets() but ignoring newlines.
 */
int affy_readchars(FILE * input, char *buf, size_t numbytes)
{
  size_t i;
  int next;

  assert(input != NULL);
  assert(buf != NULL);

  for (i = 0; i < (numbytes - 1); i++)
  {
    next = fgetc(input);

    if (next == EOF)
      return (-1);
    else
      buf[i] = next;
  }

  buf[i] = '\0';

  return (0);
}

/* 
 * affy_write8(): Write an unsigned byte to the file stream. 
 *
 * Inputs: *input is an open file input stream, *buf points to allocated
 *          storage large enough to hold an unsigned byte.
 * Outputs: 0 on success, -1 on failure (error or EOF).
 * Side effects: None.
 */
int affy_write8(FILE * input, void *buf)
{
  assert(input != NULL);
  assert(buf != NULL);

  if (fwrite(buf, 1, 1, input) < 1)
    return (-1);

  /* For bytes, that's it. */
  return (0);
}

/* 
 * affy_write16(): Write a 16-bit integer to the file stream. 
 *
 * Inputs: *input is an open file input stream, *buf points to allocated
 *          storage large enough to hold two bytes.
 * Outputs: 0 on success, -1 on failure (error or EOF).
 * Side effects: None.
 */
int affy_write16(FILE * input, void *buf)
{
  assert(input != NULL);
  assert(buf != NULL);

  AFFY_SWAP16(buf);

  if (fwrite(buf, 2, 1, input) < 1)
    return (-1);

  return (0);
}


/* 
 * affy_write32(): Write a 32-bit integer or float to the file stream. 
 *
 * Inputs: *input is an open file input stream, *buf points to allocated
 *          storage large enough to hold four bytes.
 * Outputs: 0 on success, -1 on failure (error or EOF).
 * Side effects: None.
 */
int affy_write32(FILE * input, void *buf)
{
  assert(input != NULL);
  assert(buf != NULL);

  AFFY_SWAP32(buf);

  if (fwrite(buf, 4, 1, input) < 1)
    return (-1);

  return (0);
}

/* 
 * affy_write64(): Write a 64-bit integer or float to the file stream. 
 *
 * Inputs: *input is an open file input stream, *buf points to allocated
 *          storage large enough to hold eight bytes.
 * Outputs: 0 on success, -1 on failure (error or EOF).
 * Side effects: None.
 */
int affy_write64(FILE * input, void *buf)
{
  assert(input != NULL);
  assert(buf != NULL);

  AFFY_SWAP64(buf);

  if (fwrite(buf, 8, 1, input) < 1)
    return (-1);

  return (0);
}

/* 
 * affy_writechars(): Write a series of characters to the file stream. 
 *
 * Inputs: *input is an open file input stream, *buf points to allocated
 *          storage large enough to hold the desired number of bytes.
 *          At most numbytes characters will be written
 * Outputs: 0 on success, -1 on failure (error or EOF).
 * Side effects: None.
 *
 * Comments: Provided for ease of reading character strings.  Almost
 *           equivalent to fgets() but ignoring newlines.
 */
int affy_writechars(FILE * input, char *buf, size_t numbytes)
{
  size_t i;
  int next;

  assert(input != NULL);
  assert(buf != NULL);


  if (fwrite(buf, 1, numbytes, input) < 1)
    return (-1);

  return (0);
}



int affy_readint(FILE *input) {
	int buf;
	affy_read32(input, (void *)&buf);
	return buf;
}

int affy_readubyte(FILE *input) {
	affy_uint8 buf;
	affy_read8(input, (void *)&buf);
	return (int)(unsigned char)buf;
}
int affy_readbyte(FILE *input) {
	affy_uint8 buf;
	affy_read8(input, (void *)&buf);
	return (int)(char)buf;
}

int affy_readushort(FILE *input) {
	affy_int16 buf;
	affy_read16(input, (void *)&buf);
	return (int)buf;
}
int affy_readshort(FILE *input) {
	affy_int16 buf;
	affy_read16(input, (void *)&buf);
	return (int)buf;
}

float affy_readfloat(FILE *input) {
	float buf;
	affy_read32(input, (void *)&buf);
	return buf;
}
double affy_readdouble(FILE *input) {
	double buf;
	affy_read64(input, (void *)&buf);
	return buf;
}

int affy_writeint(FILE *input, int buf) {
	return affy_write32(input, (void *)&buf);
}

int affy_writefloat(FILE *input, float buf) {
	return affy_write32(input, (void *)&buf);
}
int affy_writedouble(FILE *input, double buf) {
	return affy_write64(input, (void *)&buf);
}
