
/**************************************************************************
 *
 * Filename: intermediate_results.c
 *
 * Purpose: Save intermediate results to a file
 *
 * Creation date: 
 *
 * Author: Debashis Sahoo
 *
 *
 * Update History
 * --------------
 * 5/8/07: Created
 *
 * $Id: intermediate_results.c,v 1.4 2007/05/10 07:56:07 sahoo Exp $
 *
 **************************************************************************/

#include <affy.h>
#include <utils.h>

/* For stat system call */
#include <sys/stat.h>

#define FILENAME_SIZE 128
#define MAGIC_INTERMEDIATE 0xaa
#define NAMESIZE (FILENAME_SIZE/sizeof(int))

/*************  Hashtable ********************/

static unsigned long sdbm(unsigned char *str) {
  unsigned long hash = 0;
  int c;

  while (c = *str++)
    hash = c + (hash << 6) + (hash << 16) - hash;

  return hash;
}

static unsigned int hashfromkey(void *ky) {
  unsigned long h = sdbm( (unsigned char *) ky);
  return (unsigned int) h;
}

static int equalkeys(void *ky1, void *ky2) {
  return strcmp( (char*) ky1, (char*)ky2 ) == 0;
}

/*************  END Hashtable ********************/

void save_pm_to_disk(AFFY_CHIPSET *c, int curNum, int max_chips, char *file) {

  FILE *fp;
  struct stat bufstat;
  AFFY_CDFFILE *cdf = c->cdf;
  int numchips = c->numchips;
  int ps, i, j;
  long curPos = (2 + NAMESIZE + curNum * (NAMESIZE + cdf->numprobes)) * sizeof(int);
  long finalPos = curPos + (NAMESIZE + cdf->numprobes)* sizeof(int);
  int count = 0;
  char chipfilename[256];
  char chiptype[256];

  if (stat(file, &bufstat) < 0) {
    fp = FOPEN(file, "w+b");
  }
  else {
    fp = FOPEN(file, "r+b");
  }
  setvbuf(fp, NULL, _IOFBF, MAXBUF);

  info(("Writing to %s : %ld-%ld..\n", file, curPos, finalPos));
  affy_writeint(fp, MAGIC_INTERMEDIATE);
  affy_writeint(fp, numchips);
  strcpy(chiptype, cdf->array_type);
  affy_writechars(fp, chiptype, FILENAME_SIZE);
  if (fseek(fp, curPos, SEEK_SET) < 0) {
    info(("Error seeking to pos %ld\n", curPos));
    exit(1);
  }
  strcpy(chipfilename, c->chip[curNum]->filename);
  affy_writechars(fp, chipfilename, FILENAME_SIZE);
  for (ps = 0; ps < cdf->numprobesets; ps++) {
    AFFY_PROBESET *p = &(cdf->probeset[ps]);
    int numprobes = p->numprobes;
    count += numprobes;
    /* Load probes from pm array or actual cel data */
    for (j = 0; j < numprobes; j++) {
      int val = (int) c->chip[curNum]->pm[p->probe[j].index];
      affy_writeint(fp, val);
    } /* end probe */
  } /* end ps */
  assert(count == cdf->numprobes);
  fclose(fp);
}

double * load_pm_from_disk(AFFY_CHIPSET *c, int curNum, int max_chips, char *file) {

  struct stat bufstat;
  AFFY_CDFFILE *cdf = c->cdf;
  int numchips = c->numchips;
  int ps, i, j;
  long curPos = (2 + NAMESIZE + curNum * (NAMESIZE + cdf->numprobes)) * sizeof(int);

  double *pm = NULL;
  FILE * fp = FOPEN(file, "r+b");
  setvbuf(fp, NULL, _IOFBF, MAXBUF);

  info(("Loading pm for %d from %s..\n", curNum, file));
  if (fseek(fp, curPos+FILENAME_SIZE, SEEK_SET) < 0) {
    info(("Error seeking to pos %ld\n", curPos));
    exit(1);
  }
  pm = CALLOC(cdf->numprobes, sizeof(double));
  for (ps = 0; ps < cdf->numprobesets; ps++) {
    AFFY_PROBESET *p = &(cdf->probeset[ps]);
    int numprobes = p->numprobes;
    /* Load probes from pm array or actual cel data */
    for (j = 0; j < numprobes; j++) {
      pm[p->probe[j].index] = affy_readint(fp);
    } /* end probe */
  } /* end ps */
  fclose(fp);
  return pm;
}

int init_probeset_loading(char *file, AFFY_CHIPSET *c, AFFY_PSLOAD *ps) {
  struct stat bufstat;
  char chiptype[MAXBUF];
  int i;
  ps->fp = NULL;
  ps->chiptype = NULL;
  ps->reader = NULL;
  if (stat(file, &bufstat) < 0) {
    return -1;
  }
  ps->fp = FOPEN(file, "rb");
  ps->chipset = c;
  ps->h_probesets = NULL;
  setvbuf(ps->fp, NULL, _IOFBF, MAXBUF);
  int magic  = affy_readint(ps->fp); /* Read magic */
  if (magic != MAGIC_INTERMEDIATE) {
    return -1;
  }
  ps->max_chips = affy_readint(ps->fp); /* Read max_chips */
  ps->numchips = ps->max_chips;
  if (ps->numchips <= 0) {
    return -1;
  }
  affy_readchars(ps->fp, chiptype, FILENAME_SIZE);
  ps->chiptype = strdup(chiptype);
  ps->reader = CALLOC(ps->max_chips, sizeof(buffered_reader*));
  for (i = 0; i < ps->max_chips; i++) {
    ps->reader[i] = init_buffered_reader(ps->fp, MAXBUF*10);
  }
  return 0;
}

void close_probeset_loading(AFFY_PSLOAD *ps) {
  int i;
  if (ps->reader) {
    for (i = 0; i < ps->max_chips; i++) {
      close_buffered_reader(ps->reader[i]);
    }
    free(ps->reader);
  }
  if (ps->fp)
    fclose(ps->fp);
  if (ps->chiptype)
    free(ps->chiptype);
}

char * load_filename_from_disk(AFFY_PSLOAD *psload, int curNum) {
  AFFY_CHIPSET *c = psload->chipset;
  AFFY_CDFFILE *cdf = c->cdf;
  char filename[MAXBUF];
  int k;
  long curPos;
  curPos = (2 + NAMESIZE + curNum * (NAMESIZE + cdf->numprobes)) * sizeof(int);
  if (fseek(psload->fp, curPos, SEEK_SET) < 0) {
    info(("Error seeking to pos %ld\n", curPos));
    exit(1);
  }
  affy_readchars(psload->fp, filename, FILENAME_SIZE);
  return strdup(filename);
}

struct hashtable * get_filenames(AFFY_PSLOAD *psload) {
  struct hashtable * h;
  int i =0;
  h = create_hashtable(16, hashfromkey, equalkeys);
  for ( i =0; i < psload->numchips; i++ ) {
    char * k = load_filename_from_disk(psload, i);
    int * v = (int*) 1;
    info(("Found image %s\n", k));
    if (!hashtable_insert(h,k,v)) exit(-1); /*oom*/
  }
  return h;
}

struct hashtable * get_probesets(AFFY_CDFFILE *cdf) {
  struct hashtable * h;
  int i =0, ps = 0;
  h = create_hashtable(16, hashfromkey, equalkeys);
  for (ps = 0; ps < cdf->numprobesets; ps++)
  {
    AFFY_PROBESET *p = &(cdf->probeset[ps]);
    char * k = p->name;
    int * v = (int*) ps;
    if (!hashtable_insert(h,k,v)) exit(-1); /*oom*/
  }
  return h;
}

int readint_buffered(buffered_reader *reader) {
  char buf[8];
  int res = 0;
  read_buffered(reader, buf, 0, sizeof(int));
  affy_read32_buffer(buf, &res);
  return res;
}

void load_matrix_from_disk(AFFY_PSLOAD *psload, int ps, double **z) {
  AFFY_CHIPSET *c = psload->chipset;
  AFFY_CDFFILE *cdf = c->cdf;
  int numchips = psload->max_chips;
  double *mean = c->mean;
  int i, j, k;
  double LOG2 = log(2.0);

  AFFY_PROBESET *p = &(cdf->probeset[ps]);
  int numprobes = p->numprobes;
  int offset = 0;
  for (k = 0; k < ps; k++) {
    AFFY_PROBESET *p = &(cdf->probeset[k]);
    offset += p->numprobes;
  }
  for (i = 0; i < numchips; i++) {
    long curPos = (2 + NAMESIZE + i * (NAMESIZE + cdf->numprobes)) * sizeof(int);
    curPos += offset * sizeof(int) + FILENAME_SIZE;
#if 0
    /* No buffering */
    if (fseek(psload->fp, curPos, SEEK_SET) < 0) {
      info(("Error seeking to pos %ld\n", curPos));
      exit(1);
    }
    /* Load probes from pm array or actual cel data */
    for (j = 0; j < numprobes; j++) {
      int val = affy_readint(psload->fp);
      z[j][i] = log(mean[val]) / LOG2;
    } /* end probe */
#else
    /* Buffering during the readint */
    if (seek_buffered(psload->reader[i], curPos) < 0) {
      info(("Error seeking to pos %ld\n", curPos));
      exit(1);
    }
    /* Load probes from pm array or actual cel data */
    for (j = 0; j < numprobes; j++) {
      int val;
      val = readint_buffered(psload->reader[i]);
      z[j][i] = log(mean[val]) / LOG2;
    } /* end probe */
#endif
  } /* end chip */

}

void save_mean_to_disk(double *mean, int numprobes, char *file) {

  FILE *fp;
  char filename[256];
  int i;

  sprintf(filename, "%s.mean", file);
  fp = FOPEN(filename, "wb");
  setvbuf(fp, NULL, _IOFBF, MAXBUF);
  affy_writeint(fp, numprobes);
  for (i =0; i < numprobes; i++) {
    affy_writedouble(fp, mean[i]);
  }
  fclose(fp);
}

void load_mean_from_disk(double *mean, int *numprobes, char *file) {

  FILE *fp;
  char filename[256];
  int i;

  sprintf(filename, "%s.mean", file);
  fp = FOPEN(filename, "rb");
  setvbuf(fp, NULL, _IOFBF, MAXBUF);
  numprobes[0] = affy_readint(fp);
  for (i =0; i < numprobes[0]; i++) {
    mean[i] = affy_readdouble(fp);
  }
  fclose(fp);
}

