/**************************************************************************
 *
 * Filename: load_agcc_cel_file.c
 *
 * Purpose: Parse a AGCC CEL file and initialize an accompanying structure.
 *
 * Creation date: 6/29/2009
 *
 * Author: Debashis Sahoo
 *
 *
 **************************************************************************/

#include <affy.h>
#include <utils.h>

static void process_cel_section();
static void process_header_section();
static void process_intensity_section();
static void process_mask_section();
static void process_outlier_section();


static int pb;

#ifdef LIBUTILS_BIG_ENDIAN
# define AFFY_BIG32(x)
# define AFFY_BIG16(x)
#else
# define AFFY_BIG32(x) *(affy_uint32 *)(x) =			       \
                         ((*(affy_uint32 *)(x) << 24) |	      	       \
                         ((*(affy_uint32 *)(x) << 8) & 0x00ff0000UL) | \
                         ((*(affy_uint32 *)(x) >> 8) & 0x0000ff00UL) | \
			 (*(affy_uint32 *)(x) >> 24))
# define AFFY_BIG16(x) *(affy_uint16 *)(x) =		               \
                         (((*(affy_uint16 *)(x)) & 0x00ff) << 8) |     \
			 (((*(affy_uint16 *)(x)) >> 8))
#endif

void affy_load_agcc_cel_file(FILE *fp, AFFY_CELFILE *cf, int cf_pb)
{
  char *str;
	
	pb=cf_pb;
	/* Allocate a big buffer for reading */	
  setvbuf(fp, NULL, _IOFBF, MAXBUF);		

  /* Process by section */
	/* Magic Number */
	if ( affy_readubyte(fp) != AFFY_CEL_AGCC_MAGIC ) 
		die(("Cannot get correct magic number for binary CEL file!"));
  process_cel_section(fp, cf);
  process_header_section(fp, cf);
  process_intensity_section(fp, cf);
	
  return;
}

/*
*/
static void process_cel_section(FILE * fp, AFFY_CELFILE *cf)
{
  info(("   CEL version:  AGCC %d\n", affy_readubyte(fp)));

  return;
}

static char * affy_read_string(FILE *fp, int *size) {
  char *ptr = NULL;
  int i = 0;
  int n = 0;
  n = affy_readint(fp);
  AFFY_BIG32(&n);
  size[0] = n+1;
  if ( n < 0 ) {
    info((" Error in reading string \n"));
    exit(0);
  }
  ptr = (char*) malloc((n+1) * sizeof(char));
  for (i = 0; i < n; i++) {
    int c = fgetc(fp);
    ptr[i] = (char) (c & 0xff);
    if (c == EOF) break;
  }
  ptr[n] = '\0';
  return ptr;
}

static char * affy_read_wstring1(FILE *fp, int *size) {
  char *ptr = NULL;
  int i = 0;
  int n = 0;
  n = affy_readint(fp);
  AFFY_BIG32(&n);
  size[0] = n+1;
  if ( n < 0 ) {
    info((" Error in reading wstring1 \n"));
    exit(0);
  }
  ptr = (char*) malloc((n+1) * sizeof(char));
  for (i = 0; i < n; i++) {
    int c = affy_readshort(fp);
    affy_int16 k = (affy_int16) (c & 0xffff);
    AFFY_BIG16(&k);
    ptr[i] = (char) (k & 0xff);
    if (c == EOF) break;
  }
  ptr[n] = '\0';
  return ptr;
}

static affy_int16 * affy_read_wstring(FILE *fp, int *size) {
  affy_int16 *ptr = NULL;
  int i = 0;
  int n = 0;
  n = affy_readint(fp);
  AFFY_BIG32(&n);
  size[0] = n+1;
  if ( n < 0 ) {
    info((" Error in reading wstring \n"));
    exit(0);
  }
  ptr = (affy_int16*) malloc((n+1) * sizeof(affy_int16));
  for (i = 0; i < n; i++) {
    int c = affy_readshort(fp);
    ptr[i] = (affy_int16) (c & 0xffff);
    AFFY_BIG16(&ptr[i]);
    if (c == EOF) break;
  }
  ptr[n] = (affy_int16) 0;
  return ptr;
}

static char * wstring_to_string(char *ptr, int size) {
  int n = size/2;
  int i =0;
  char * str = (char*)malloc(n*sizeof(char));
  for (i = 0; i < n; i++) {
    str[i] = ptr[i*2+1];
  }
  return str;
}

static int affy_read_int(FILE * fp) {
  int n = affy_readint(fp);
  AFFY_BIG32(&n);
  return n;
}

static affy_float32 affy_read_float(FILE * fp) {
  affy_float32 res;
  int val = affy_readint(fp);
  AFFY_BIG32(&val);
  memcpy(&res, &val, sizeof(affy_int32));
  return res;
}

static affy_int16 affy_read_short(FILE * fp) {
  int c = affy_readshort(fp);
  affy_int16 k = (affy_int16) (c & 0xffff);
  AFFY_BIG16(&k);
  return k;
}

void affy_agcc_get_cdf_buf(FILE * fp, char *buf, int size)
{
  char * hdr;
  int length, n, i;
  /* 4 bytes already read */
  affy_readshort(fp);
  /* Now File position of the first data group.*/
  affy_readint(fp);
  do {
    hdr = affy_read_string(fp, &length);
    // printf("%d: Id %s :-\n", length, hdr);
    free(hdr);
    hdr = affy_read_string(fp, &length);
    // printf("%d: Guid %s :-\n", length, hdr);
    free(hdr);
    hdr = affy_read_wstring1(fp, &length);
    // printf("%d: time %s :-\n", length, hdr);
    free(hdr);
    hdr = affy_read_wstring1(fp, &length);
    // printf("%d: locale %s :-\n", length, hdr);
    free(hdr);
    n = affy_read_int(fp);
    // printf("Num par = %d :-\n", n);
    for (i = 0; i < n; i++) {
      char *name, *value, *type;
      int size_v;
      name = affy_read_wstring1(fp, &length);
      // printf("%d: %s :-%d\n", length, name, i);
      value = affy_read_string(fp, &size_v);
      type = affy_read_wstring1(fp, &length);
      // printf("%d: %s :-%d\n", length, type, i);
      if (strcmp(type, "text/plain") == 0) {
	char * val = wstring_to_string(value, size_v);
	// printf("%s :-%d\n", val, i);
	if (strcmp(name, "affymetrix-partial-dat-header") == 0) {
	  if (size_v < size) {
	    size = size_v;
	  }
	  strncpy(buf, val, size);
	}
	free(val);
      }
      free(name);
      free(value);
      free(type);
    }
    n = affy_read_int(fp);
    // printf("Num parent headers = %d :-\n", n);
  } while (n != 0);
}

static void process_header_section(FILE * fp, AFFY_CELFILE *cf)
{
  int i, n;
  int total;
  char *hdr;
  int numDataGroup;
  int posDataGroup;
  int locale;
  int length;
  long fptr;
	
  cf->numrows = -1;
  cf->numcols = -1;
  numDataGroup = affy_read_int(fp);
  posDataGroup = affy_read_int(fp);
  do {
    hdr = affy_read_string(fp, &length);
    /* printf("%d: Id %s :-\n", length, hdr); */
    free(hdr);
    hdr = affy_read_string(fp, &length);
    /* printf("%d: Guid %s :-\n", length, hdr); */
    free(hdr);
    hdr = affy_read_wstring1(fp, &length);
    /* printf("%d: time %s :-\n", length, hdr); */
    free(hdr);
    hdr = affy_read_wstring1(fp, &length);
    /* printf("%d: locale %s :-\n", length, hdr); */
    free(hdr);
    n = affy_read_int(fp);
    /* printf("Num par = %d :-\n", n); */
    for (i = 0; i < n; i++) {
      char *name, *value, *type;
      int size_v;
      name = affy_read_wstring1(fp, &length);
      /* printf("%d: %s :-%d\n", length, name, i); */
      value = affy_read_string(fp, &size_v);
      type = affy_read_wstring1(fp, &length);
      /* printf("%d: %s :-%d\n", length, type, i); */
      if (strcmp(type, "text/x-calvin-integer-32") == 0) {
	affy_int32 val;
	memcpy(&val, value, sizeof(affy_int32));
	val=(((val>>24)&0xff) | ((val&0xff)<<24) |
	    ((val>>8)&0xff00) | ((val&0xff00)<<8));  
	/* printf("%d :-%d\n", val, i); */
	if (strcmp(name, "affymetrix-cel-rows") == 0) {
	  cf->numrows = val;
	}
	if (strcmp(name, "affymetrix-cel-cols") == 0) {
	  cf->numcols = val;
	}
      }
      if (strcmp(type, "text/x-calvin-float") == 0) {
	affy_int32 val;
	float res;
	memcpy(&val, value, sizeof(affy_int32));
	val=(((val>>24)&0xff) | ((val&0xff)<<24) |
	    ((val>>8)&0xff00) | ((val&0xff00)<<8));  
	memcpy(&res, &val, sizeof(affy_int32));
	/* printf("%f :-%d\n", res, i); */
      }
      if (strcmp(type, "text/plain") == 0) {
	char * val = wstring_to_string(value, size_v);
	/* printf("%s :-%d\n", val, i); */
	free(val);
      }
      free(name);
      free(value);
      free(type);
    }
    n = affy_read_int(fp);
    /* printf("Num parent headers = %d :-\n", n); */
  } while (n != 0);
  if (cf->numrows < 0 || cf->numcols < 0) {
    info((" Error in reading numrows (%d) and numcols (%d)\n", cf->numrows, cf->numcols));
    exit(0);
  }
  /* Once we know the parameters, we allocate storage */
  n = numbytes(cf->numcols);
  cf->data = MALLOC(cf->numrows * sizeof(AFFY_CELL *));
  cf->data[0] = CALLOC(cf->numcols * cf->numrows, sizeof(AFFY_CELL));
  cf->mask = MALLOC(cf->numrows * sizeof(affy_uint8 *));
  cf->mask[0] = CALLOC(n * cf->numrows, sizeof(affy_uint8));
  cf->outlier = MALLOC(cf->numrows * sizeof(affy_uint8 *));
  cf->outlier[0] = CALLOC(n * cf->numrows, sizeof(affy_uint8));

  for (i = 1; i < cf->numrows; i++)
  {
    cf->data[i] = cf->data[0] + (i * cf->numcols);
    cf->mask[i] = cf->mask[0] + (i * n);
    cf->outlier[i] = cf->outlier[0] + (i * n);
  }

  /* DS: To be filled */
  cf->numoutliers=0;
  cf->nummasks=0;
	
  info(("   Dim: %dx%d\n", cf->numrows, cf->numcols));
  
  fptr = ftell(fp);
  if (fptr < posDataGroup && fseek(fp, posDataGroup - fptr, SEEK_CUR) < 0) {
    info(("   Error in setting file position\n"));
    exit(0);
  }
  fptr = ftell(fp);
  if (fptr != posDataGroup) {
    info(("   Error in file position %ld != %d\n", fptr, posDataGroup));
    exit(0);
  }
  
  return;
}

static void process_intensity_section(FILE * fp, AFFY_CELFILE *cf)
{
  int x, y, i;
  int num_read = 0, progress, incr;
  int num_cells=cf->numrows*cf->numcols;
  int numDataSets = 0;
  int posNextSets = 0;
  int dataset;
  char * name;
  int length;
  /* next datagroup == 0 */
  if (affy_read_int(fp) != 0) {
    info(("   Multiple data group not allowed\n"));
    exit(0);
  }
  posNextSets = affy_read_int(fp);
  numDataSets = affy_read_int(fp);
  name = affy_read_wstring1(fp, &length);
  free(name);

  for (dataset = 0; dataset < numDataSets; dataset++) {
    long fptr = ftell(fp);
    int firstDataElement, n;
    if (fptr < posNextSets && fseek(fp, posNextSets - fptr, SEEK_CUR) < 0) {
      info(("   Error in setting file position\n"));
      exit(0);
    }
    fptr = ftell(fp);
    if (fptr != posNextSets) {
      info(("   Error in file position %ld != %d\n", fptr, posNextSets));
      exit(0);
    }
    firstDataElement = affy_read_int(fp);
    posNextSets = affy_read_int(fp);
    name = affy_read_wstring1(fp, &length);
    /* printf("Name: %s\n", name); */
    n = affy_read_int(fp);
    for (i = 0; i < n; i++) {
      char *name, *value, *type;
      int size_v;
      name = affy_read_wstring1(fp, &length);
      value = affy_read_string(fp, &size_v);
      type = affy_read_wstring1(fp, &length);
      free(name);
      free(value);
      free(type);
    }
    if (strcmp(name, "Intensity") == 0) {
      incr = num_cells / 100;
      progress = incr;

      pb_update(pb, 1, "Loading intensities");

      x=0;
      y=0;
	
      n = affy_read_int(fp);
      if (n != 1) {
	info(("   Error in Intensity column size\n"));
	exit(0);
      }
      for (i = 0; i < n; i++) {
	char *str; 
	int type, size;
	str = affy_read_wstring1(fp, &length);
	type = affy_readbyte(fp);
	size = affy_read_int(fp);
	if (type != 6) {
	  info(("   Error in Intensity type\n"));
	  exit(0);
	}
	free(str);
      }
      n = affy_read_int(fp);
      if (n != num_cells) {
	info(("   Incorrect num cells \n"));
	exit(0);
      }
      for (i = 0; i < n; i++) {
	affy_float32 res;
	if (i == progress) {
	  progress += incr;
	  pb_update(pb, progress / incr, "");
	}
	/* Check for next row */
	if ( x == cf->numcols ) {
	  x=0;
	  y++;
	}
	res = affy_read_float(fp);
	cf->data[x][y].value = res;
	bset(cf->mask[x], y, 1);
	bset(cf->outlier[x], y, 1);
	x++;
      }
    }
    if (strcmp(name, "StdDev") == 0) {
      incr = num_cells / 100;
      progress = incr;

      pb_update(pb, 1, "Loading StdDev");

      x=0;
      y=0;
	
      n = affy_read_int(fp);
      if (n != 1) {
	info(("   Error in StdDev column size\n"));
	exit(0);
      }
      for (i = 0; i < n; i++) {
	char *str; 
	int type, size;
	str = affy_read_wstring1(fp, &length);
	type = affy_readbyte(fp);
	size = affy_read_int(fp);
	if (type != 6) {
	  info(("   Error in StdDev type\n"));
	  exit(0);
	}
	free(str);
      }
      n = affy_read_int(fp);
      if (n != num_cells) {
	info(("   Incorrect num cells \n"));
	exit(0);
      }
      for (i = 0; i < n; i++) {
	affy_float32 res;
	if (i == progress) {
	  progress += incr;
	  pb_update(pb, progress / incr, "");
	}
	/* Check for next row */
	if ( x == cf->numcols ) {
	  x=0;
	  y++;
	}
	res = affy_read_float(fp);
	cf->data[x][y].stddev = res;
	x++;
      }
    }
    if (strcmp(name, "Pixel") == 0) {
      incr = num_cells / 100;
      progress = incr;

      pb_update(pb, 1, "Loading Pixel");

      x=0;
      y=0;
	
      n = affy_read_int(fp);
      if (n != 1) {
	info(("   Error in Pixel column size\n"));
	exit(0);
      }
      for (i = 0; i < n; i++) {
	char *str; 
	int type, size;
	str = affy_read_wstring1(fp, &length);
	type = affy_readbyte(fp);
	size = affy_read_int(fp);
	if (type != 2) {
	  info(("   Error in Pixel type\n"));
	  exit(0);
	}
	free(str);
      }
      n = affy_read_int(fp);
      if (n != num_cells) {
	info(("   Incorrect num cells \n"));
	exit(0);
      }
      for (i = 0; i < n; i++) {
	affy_int16 res;
	if (i == progress) {
	  progress += incr;
	  pb_update(pb, progress / incr, "");
	}
	/* Check for next row */
	if ( x == cf->numcols ) {
	  x=0;
	  y++;
	}
	res = affy_read_short(fp);
	cf->data[x][y].numpixels = res;
	x++;
      }
    }
    if (strcmp(name, "Outlier") == 0) {
      n = affy_read_int(fp);
      if (n != 2) {
	info(("   Error in Outlier column size\n"));
	exit(0);
      }
      for (i = 0; i < n; i++) {
	char *str; 
	int type, size;
	str = affy_read_wstring1(fp, &length);
	type = affy_readbyte(fp);
	size = affy_read_int(fp);
	if (type != 2) {
	  info(("   Error in Outlier type\n"));
	  exit(0);
	}
	free(str);
      }
      n = affy_read_int(fp);
      /* printf("Num = %d\n", n); */
      cf->numoutliers=n;
      for (i = 0; i < n; i++) {
	affy_int16 x, y;
	x = affy_read_short(fp);
	y = affy_read_short(fp);
	/* printf("[%d, %d]:[%d/%d]\n", x, y, i, n); */
	bset(cf->outlier[x], y, 0);
      }
    }
    if (strcmp(name, "Mask") == 0) {
      n = affy_read_int(fp);
      if (n != 2) {
	info(("   Error in Mask column size\n"));
	exit(0);
      }
      for (i = 0; i < n; i++) {
	char *str; 
	int type, size;
	str = affy_read_wstring1(fp, &length);
	type = affy_readbyte(fp);
	size = affy_read_int(fp);
	if (type != 2) {
	  info(("   Error in Mask type\n"));
	  exit(0);
	}
	free(str);
      }
      n = affy_read_int(fp);
      /* printf("Num = %d\n", n); */
      cf->nummasks=n;
      for (i = 0; i < n; i++) {
	affy_int16 x, y;
	x = affy_read_short(fp);
	y = affy_read_short(fp);
	/* printf("[%d, %d]:[%d/%d]\n", x, y, i, n); */
	bset(cf->mask[x], y, 0);
      }
    }
    free(name);
  }
  return;
}

