
/**************************************************************************
 *
 * Filename: load_binary_cdf_file.c
 *
 * Purpose: Parse a binary CDF file and initialize an accompanying structure.
 *
 * Creation date: 11-10-2005
 *
 * Author: Steven Eschrich
 *
 *
 * Update History
 * --------------
 * 4/8/05: Imported/repaired from old libaffy (AMH)
 *
 **************************************************************************/

#include <affy.h>
#include <utils.h>

/* XXX -- remove access() calls and handle in some platform-independent way */
#include <unistd.h>

static void process_chip_section();
static void process_qc_section();
static void process_probe_section();

/*
  The CDF file contains the description of the microarray chip in terms
  of probes and probe sets. This information needs to be loaded into
  memory so that calculations at the probe level can be matched
  to (X,Y) coordinates in the CEL file.
  
  The parameter is a chip_type as defined by the CEL file. This should
  correspond to a cdf file somewhere.
*/
void affy_load_binary_cdf_file(FILE *fp, AFFY_CDFFILE *cdf, int cdf_pb)
{
  int i;
  int progress = 0, probe_incr = 0;
	char *ps_name;

  setvbuf(fp, NULL, _IOFBF, MAXBUF);
	/* This file is in a particular order and so is rather linearly parsed here. */

        /* Skip the magic number */
        affy_readint(fp);

	/* Version */
	info(("  CDF version %d\n",affy_readint(fp)));
	/* Num columns/rows */
	cdf->numcols=affy_readushort(fp);
	cdf->numrows=affy_readushort(fp);
	cdf->numprobesets=affy_readint(fp);
	cdf->numqcunits=affy_readint(fp);

  /* At this point we know the size of the chip, and the number
     of probes.
   */
  cdf->cell_type = MALLOC(cdf->numrows * sizeof(BYTE *));
  cdf->xy_ref = MALLOC(cdf->numrows * sizeof(AFFY_PROBE **));
  for (i = 0; i < cdf->numrows; i++)
  {
    cdf->cell_type[i] = CALLOC(cdf->numcols, sizeof(BYTE));
    cdf->xy_ref[i] = CALLOC(cdf->numcols, sizeof(AFFY_PROBE *));
  }
  cdf->probeset = CALLOC(cdf->numprobesets, sizeof(AFFY_PROBESET));
  cdf->probe =
    CALLOC(cdf->numrows * cdf->numcols / 2.0, sizeof(AFFY_PROBE *));

	/* Resequencing reference sequence, ignore */
	fseek(fp, affy_readint(fp), SEEK_CUR);
	
	/* Probeset names */
	ps_name=CALLOC(65,sizeof(char));
	for (i=0; i < cdf->numprobesets; i++) {
		affy_readchars(fp,ps_name,65);
		cdf->probeset[i].name=strdup(ps_name);
	}
	free(ps_name);
	
	/* Skip details about file positions */
	fseek(fp, cdf->numqcunits*4+cdf->numprobesets*4,SEEK_CUR);		
			
	probe_incr = cdf->numprobesets / 100;
  progress = probe_incr;
  pb_update(cdf_pb, 1, "Loading probes");
	
	/* Process QC section */
  process_qc_section(fp, cdf);	

	/* Process Probe section */
	for (i=0; i < cdf->numprobesets; i++) {
	  if (i == progress) {
        pb_update(cdf_pb, progress / probe_incr, "");
        progress += probe_incr;
    }
    process_probe_section(fp, cdf, i);
	}
	
  return;
}

/* A quality control section */
static void process_qc_section(FILE * fp, AFFY_CDFFILE *cdf)
{
  int num_qc_cells = 0, qc_counter;
  int x, y;
	int i;
	
	/* For each qc unit, read in details */
	for (i=0; i < cdf->numqcunits; i++) {
		/* Get (don't use) the type */
		affy_readushort(fp);
		/* Number of cells in qc unit */
		num_qc_cells=affy_readint(fp);
		for (qc_counter=0; qc_counter < num_qc_cells; qc_counter++) {
			x=affy_readushort(fp);
			y=affy_readushort(fp);
			cdf->cell_type[x][y]=AFFY_QC_LOCATION;
			cdf->xy_ref[x][y]=NULL;
			/* Skip Probe length, Perfect Match flag, Background probe flag */
			fseek(fp,3,SEEK_CUR);
		}

	}
	return;
}

/* A probe section will be indexed by probe, and by x,y coordinates */
static void process_probe_section(FILE * fp, AFFY_CDFFILE *cdf, int ps)
{
  int status = AFFY_NORMAL_LOCATION;
  int numprobes = 0;
  char **kv, **f;
  char *s;
  int x, y, i, j;
  char pbase, tbase, atom;


	/* Guess status */
  if (!strncmp(cdf->probeset[ps].name, "AFFX", 4)) {
	  status = AFFY_CONTROL_LOCATION;
  }
	/* UnitType */
	fseek(fp,2,SEEK_CUR);
	/* Direction */
	fseek(fp,1,SEEK_CUR);
	/* Num atoms */
  numprobes = affy_readint(fp);
  /* Num blocks - always 1 for expressions */
  fseek(fp,4,SEEK_CUR);
  /* Number of Cells */
  fseek(fp,4,SEEK_CUR);
  /* Unit number */
  fseek(fp,4,SEEK_CUR);
  /* Number cells per atom */
  fseek(fp,1,SEEK_CUR);
  
  /* Allocate enough storage for all probes in this probe set */
  cdf->probeset[ps].probe = CALLOC(numprobes, sizeof(AFFY_PROBE));
  cdf->probeset[ps].numprobes = numprobes;
  cdf->probeset[ps].index = ps;

  /* For each block (only one block) */
  /* Num atoms */
	fseek(fp,4,SEEK_CUR);
	/* Num cells */
	fseek(fp,4,SEEK_CUR);
	/* Num cells/atom */
	fseek(fp,1,SEEK_CUR);
	/*  Direction */
	fseek(fp,1,SEEK_CUR);
	/* Position of first atom */
	fseek(fp,4,SEEK_CUR);
	/* Position of second atom */
	fseek(fp,4,SEEK_CUR);
	/* Block name */
	fseek(fp,64,SEEK_CUR);

  /* Get each probe */
  for (i = 0; i < numprobes; i++)
  {
    for (j = 0; j < 2; j++)
    {
      atom = affy_readint(fp);
      x = affy_readushort(fp);
      y = affy_readushort(fp);
      /* Index position, ignore */
      fseek(fp, 4, SEEK_CUR);
      pbase = fgetc(fp);
      tbase = fgetc(fp);


      cdf->cell_type[x][y] = status;
      if (pbase != tbase)
      {
        cdf->probeset[ps].probe[i].pm.x = x;
        cdf->probeset[ps].probe[i].pm.y = y;
      }
      else
      {
        cdf->probeset[ps].probe[i].mm.x = x;
        cdf->probeset[ps].probe[i].mm.y = y;
      }
      cdf->xy_ref[x][y] = &(cdf->probeset[ps].probe[i]);
    }
    cdf->probeset[ps].probe[i].ps = &(cdf->probeset[ps]);

    /* Add probe to list of all probes */
    cdf->probe[cdf->numprobes] = &(cdf->probeset[ps].probe[i]);
    cdf->probe[cdf->numprobes]->index = cdf->numprobes;
    cdf->numprobes++;
  }

  return;
}
