
/**************************************************************************
 *
 * Filename: load_cel_file.c
 *
 * Purpose: Parse a CEL file and initialize an accompanying structure.
 *
 * Creation date: 
 *
 * Author: Steven Eschrich
 *
 *
 * Update History
 * --------------
 * 4/8/05: Imported/repaired from old libaffy (AMH)
 * 4/19/05: Now we use AFFY_CELL and collect all cell data (AMH)
 *
 **************************************************************************/

#include <affy.h>
#include <utils.h>

static void process_cel_section();
static void process_header_section();
static void process_intensity_section();
static void process_mask_section();
static void process_outlier_section();


static int pb;


void affy_load_binary_cel_file(FILE *fp, AFFY_CELFILE *cf, int cf_pb)
{
  char *str;
	
	pb=cf_pb;
	/* Allocate a big buffer for reading */	
  setvbuf(fp, NULL, _IOFBF, MAXBUF);		

  /* Process by section */
	/* Magic Number */
	if ( affy_readint(fp) != 64 ) 
		die(("Cannot get correct magic number for binary CEL file!"));
  process_cel_section(fp, cf);
  process_header_section(fp, cf);
  process_intensity_section(fp, cf);
	process_mask_section(fp, cf);
	process_outlier_section(fp, cf);

	/* Skip the sub-grid section */
	
  return;
}

/*
*/
static void process_cel_section(FILE * fp, AFFY_CELFILE *cf)
{
  info(("   CEL version:  %d\n", affy_readint(fp)));

  return;
}

static void process_header_section(FILE * fp, AFFY_CELFILE *cf)
{
  int i, n;
	int total;
	char *hdr;
	
	cf->numcols = affy_readint(fp);
  cf->numrows = affy_readint(fp);
	/* Num cells, guess not needed */
	affy_readint(fp);
	
  /* Once we know the parameters, we allocate storage */
  n = numbytes(cf->numcols);
  cf->data = MALLOC(cf->numrows * sizeof(AFFY_CELL *));
  cf->data[0] = CALLOC(cf->numcols * cf->numrows, sizeof(AFFY_CELL));
  cf->mask = MALLOC(cf->numrows * sizeof(affy_uint8 *));
  cf->mask[0] = CALLOC(n * cf->numrows, sizeof(affy_uint8));
  cf->outlier = MALLOC(cf->numrows * sizeof(affy_uint8 *));
  cf->outlier[0] = CALLOC(n * cf->numrows, sizeof(affy_uint8));

  for (i = 1; i < cf->numrows; i++)
  {
    cf->data[i] = cf->data[0] + (i * cf->numcols);
    cf->mask[i] = cf->mask[0] + (i * n);
    cf->outlier[i] = cf->outlier[0] + (i * n);
  }

#if 0
	/* At the moment, remainder of header is ignored */
	/* Text header */
	n = affy_readint(fp);
	fseek(fp, n, SEEK_CUR);
	/* Algorithm name */
	n = affy_readint(fp);
	fseek(fp, n, SEEK_CUR);

	/* Algorithm parameters */
	n = affy_readint(fp);
	fseek(fp, n, SEEK_CUR);
#else
	n = affy_readint(fp);
	for (i = 0; i < n; i++) {
	  if (fgetc(fp) == EOF) break;
	}
	n = affy_readint(fp);
	for (i = 0; i < n; i++) {
	  if (fgetc(fp) == EOF) break;
	}
	n = affy_readint(fp);
	for (i = 0; i < n; i++) {
	  if (fgetc(fp) == EOF) break;
	}
#endif

	/* Cell margin */
	affy_readint(fp);
	
	cf->numoutliers=affy_readint(fp);
	cf->nummasks=affy_readint(fp);
	/* Num subgrids */
	affy_readint(fp);
	
  info(("   Dim: %dx%d\n", cf->numrows, cf->numcols));

  return;
}

static void process_intensity_section(FILE * fp, AFFY_CELFILE *cf)
{
  int x, y, i;
  int num_read = 0, progress, incr;
	int num_cells=cf->numrows*cf->numcols;
	affy_float32 fval;
	affy_int16 sval;
	
  incr = num_cells / 100;
  progress = incr;

  pb_update(pb, 1, "Loading intensities");

	x=0;
	y=0;
	
	for (i=0; i < num_cells; i++) {

		/* Otherwise, this is a line of x,y coordinates and mean intensity */
    if (i == progress) {
		  progress += incr;
      pb_update(pb, progress / incr, "");
    }
    /* Check for next row */
    if ( x == cf->numcols ) {
    	x=0;
    	y++;
    }

  	if ( affy_read32(fp, (void *)&fval) == -1 )
  		die(("Cannot continue, there was a problem reading the intensity from file."));

    cf->data[x][y].value = fval;

  	if ( affy_read32(fp, (void *)&fval) == -1 )
  		die(("Cannot continue, there was a problem reading the stddev from file."));
    cf->data[x][y].stddev = fval;
		if ( affy_read16(fp, (void *)&sval) == -1 )
			die(("Cannot continue. there was a problem reading the numpixels from file."));
    cf->data[x][y].numpixels = sval;
    
    bset(cf->mask[x], y, 1);
    bset(cf->outlier[x], y, 1);
		x++;
  }

  return;
}

static void process_mask_section(FILE * fp, AFFY_CELFILE *cf)
{
  int x, y;
	int i;
  pb_update(pb, 1, "Processing Masks");

  info(("(%d masks)...", cf->nummasks));
	for (i=0; i < cf->nummasks; i++) {
		x=affy_readushort(fp);
		y=affy_readushort(fp);
    bset(cf->mask[x], y, 0);
  }

  return;
}

static void process_outlier_section(FILE * fp, AFFY_CELFILE *cf)
{
  int x, y;
	int i;
	
  pb_update(pb, 1, "Processing Outliers");
  info(("(%d outliers)...", cf->numoutliers));

	for (i=0; i < cf->numoutliers; i++) {
		x=affy_readshort(fp);
		y=affy_readshort(fp);
    bset(cf->outlier[x], y, 0);
  }


  return;
}

