
/**************************************************************************
 *
 * Filename: load_dat_file.c
 * 
 * Purpose: Binary DAT file reading routines.
 *
 * Creation date: 1 April, 2005
 *
 * Author: Andrew M. Hoerter
 *
 *
 * Update History
 * --------------
 * 4/1/05: Creation (AMH)
 * 4/19/05: Now we use an AFFY_PIXREGION, plus other loose ends (AMH)
 *
 **************************************************************************/

#include <affy.h>
#include <util_log.h>

#define TMPBUF_SIZE 250
#define FIELD_START "\x14 "
#define FIELD_END   " \x14"

/* 
 * affy_load_dat_file(): read in a DAT file and fill out a new AFFY_DATFILE
 *                       structure.
 *
 * Inputs: *filename contains the desired filename.
 * Outputs: pointer to AFFY_DATFILE structure on success, NULL on failure
 */
AFFY_DATFILE *affy_load_dat_file(char *filename)
{
  AFFY_DATFILE *newdat;
  FILE *input;
  affy_uint8 magic;
  affy_uint32 i;
  char tmpbuf[TMPBUF_SIZE];
  char *walk, *save;

  assert(filename != NULL);

  if ((input = fopen(filename, "rb")) == NULL)
    return (NULL);

  newdat = (AFFY_DATFILE *)malloc(sizeof(AFFY_DATFILE));
  if (newdat == NULL)
  {
    fclose(input);
    return (NULL);
  }

  newdat->experiment_name = newdat->probe_array_type = NULL;

  newdat->pixels.data = NULL;

  /* Begin actual parsing of the file. */

  /* Check magic. */
  if (affy_read8(input, &magic) != 0)
    goto err;

  if (magic != AFFY_DAT_FILEMAGIC)
    goto err;

  if (affy_read16(input, (void *)(&newdat->pixels.numcols)) != 0)
    goto err;

  if (affy_read16(input, (void *)(&newdat->pixels.numrows)) != 0)
    goto err;

  if (affy_read32(input, (void *)(&newdat->numpixels)) != 0)
    goto err;

  if (affy_read32(input, (void *)(&newdat->minpixel)) != 0)
    goto err;

  if (affy_read32(input, (void *)(&newdat->maxpixel)) != 0)
    goto err;

  if (affy_read64(input, (void *)(&newdat->meanpixel)) != 0)
    goto err;

  if (affy_read64(input, (void *)(&newdat->std_dev_pixel)) != 0)
    goto err;

/*
  if (affy_readchars(input, tmpbuf, 10) != 0)
    goto err;
  newdat->pixels_per_row = strtoul(tmpbuf+4, NULL, 10);

  if (affy_readchars(input, tmpbuf, 10) != 0)
    goto err;
  newdat->numrows = strtoul(tmpbuf+4, NULL, 10);
*/
  fseek(input, 18, SEEK_CUR);

  if (affy_readchars(input, tmpbuf, 8) != 0)
    goto err;
  newdat->pixel_width = strtoul(tmpbuf + 4, NULL, 10);

  if (affy_readchars(input, tmpbuf, 8) != 0)
    goto err;
  newdat->pixel_height = strtoul(tmpbuf + 4, NULL, 10);

  if (affy_readchars(input, tmpbuf, 7) != 0)
    goto err;
  newdat->scanspeed = strtoul(tmpbuf + 3, NULL, 10);

  if (affy_readchars(input, tmpbuf, 8) != 0)
    goto err;
  newdat->temperature = strtof(tmpbuf, NULL);

  if (affy_readchars(input, tmpbuf, 5) != 0)
    goto err;
  newdat->laser_power = strtof(tmpbuf, NULL);

  if (affy_readchars(input, tmpbuf, sizeof(newdat->timestamp)) != 0)
    goto err;
  strcpy(newdat->timestamp, tmpbuf);

  /* Variable-size subfields, read them as one unit. */
  if (affy_readchars(input, tmpbuf, 221) != 0)
    goto err;

  if ((walk = strstr(tmpbuf, FIELD_START)) == NULL)
    goto err;
  else
    *walk++ = '\0';

  newdat->scannerid = (char *)malloc(strlen(tmpbuf) + 1);
  if (newdat->scannerid == NULL)
    goto err;

  strcpy(newdat->scannerid, tmpbuf);

  if ((walk = strstr(walk, FIELD_START)) == NULL)
    goto err;
  else
    save = walk + strlen(FIELD_START);

  if ((walk = strstr(walk, ".1sq")) == NULL)
    goto err;
  else
    *walk = '\0';

  newdat->probe_array_type = (char *)malloc(strlen(save) + 1);
  if (newdat->probe_array_type == NULL)
    goto err;

  strcpy(newdat->probe_array_type, save);

  if (affy_read64(input, (void *)(&newdat->avg_dc_offset)) != 0)
    goto err;

  if (affy_read64(input, (void *)(&newdat->std_dev_dc_offset)) != 0)
    goto err;

  if (affy_read32(input, (void *)(&newdat->numsamples_dc_offset)) != 0)
    goto err;

  if (affy_read16(input, (void *)(&newdat->grid_ul.x)) != 0)
    goto err;

  if (affy_read16(input, (void *)(&newdat->grid_ul.y)) != 0)
    goto err;

  if (affy_read16(input, (void *)(&newdat->grid_ur.x)) != 0)
    goto err;

  if (affy_read16(input, (void *)(&newdat->grid_ur.y)) != 0)
    goto err;

  if (affy_read16(input, (void *)(&newdat->grid_lr.x)) != 0)
    goto err;

  if (affy_read16(input, (void *)(&newdat->grid_lr.y)) != 0)
    goto err;

  if (affy_read16(input, (void *)(&newdat->grid_ll.x)) != 0)
    goto err;

  if (affy_read16(input, (void *)(&newdat->grid_ll.y)) != 0)
    goto err;

  if (affy_read16(input, (void *)(&newdat->cellmargin)) != 0)
    goto err;

  if (affy_readchars(input, tmpbuf, 155) != 0)
    goto err;

  newdat->experiment_name = (char *)malloc(strlen(tmpbuf) + 1);
  if (newdat->experiment_name == NULL)
    goto err;

  strcpy(newdat->experiment_name, tmpbuf);

  /* Allocate pixelmap memory. */
  newdat->pixels.data =
    (unsigned int **)malloc(sizeof(unsigned int *) *
                            newdat->pixels.numrows);
  if (newdat->pixels.data == NULL)
    goto err;

  newdat->pixels.data[0] =
    (unsigned int *)malloc(sizeof(unsigned int) * newdat->numpixels);
  if (newdat->pixels.data[0] == NULL)
    goto err;

  /* Fixup the pointers for the "lines" dimension of the pixelmap. */
  for (i = 1; i < newdat->pixels.numrows; i++)
    newdat->pixels.data[i] =
      newdat->pixels.data[i - 1] + newdat->pixels.numcols;

  for (i = 0; i < newdat->numpixels; i++)
    if (affy_read16(input, (void *)(newdat->pixels.data[0] + i)) != 0)
      goto err;

  fclose(input);
  return (newdat);

err:
  affy_free_dat_file(newdat);
  fclose(input);
  die(("Couldn't load DAT file %s!", filename));
}
