
/**************************************************************************
 *
 * Filename: load_text_cdf_file.c
 *
 * Purpose: Parse a CDF file and initialize an accompanying structure.
 *
 * Creation date: 
 *
 * Author: Steven Eschrich
 *
 *
 * Update History
 * --------------
 * 4/8/05: Imported/repaired from old libaffy (AMH)
 *
 **************************************************************************/
#include <affy.h>
#include <utils.h>


static void process_chip_section();
static void process_qc_section();
static void process_probe_section();
void affy_load_text_cdf_file(FILE *fp, AFFY_CDFFILE *cdf, int cdf_pb)
{
 
  char *str;
  int probe_set_num = 0;
  char **kv;
  int unit_no, block_no;
  int progress = 0, probe_incr = 0;
    
   /* Parse file, looking at [] section headings */
  affy_reset_next_line(fp);
  while ((str = affy_get_next_line(fp)) != NULL)
  {
    /* Process by section */
    if (STREQ(str, "[CDF]"))
    {
      str = affy_get_next_line(fp);
      kv = split(str, '=');
      if (STREQ(kv[0], "Version"))
        info(("  CDF version %s\n", kv[1]));
    }
    else if (STREQ(str, "[Chip]"))
    {
      process_chip_section(fp, cdf);
      probe_incr = cdf->numprobesets / 100;
      progress = probe_incr;
      pb_update(cdf_pb, 1, "Loading probes");
    }
    else if (!strncmp(str, "[QC", 3))
    {
      process_qc_section(fp, cdf, atoi(str + 3));
    }
    else if (sscanf(str, "[Unit%d_Block%d]", &unit_no, &block_no) == 2)
    {
      if (probe_set_num == progress)
      {
        pb_update(cdf_pb, progress / probe_incr, "");
        progress += probe_incr;
      }
      process_probe_section(fp, cdf, probe_set_num++);
    }
    else
    {
      affy_skip_to_next_header(fp);
    }
  }

  pb_finish(cdf_pb);

  return;
}

static void process_chip_section(FILE * fp, AFFY_CDFFILE *cdf)
{
  char *s;
  char **kv;
  int i;

  /* Read lines until eof or a new section */
  while ((s = affy_get_next_line(fp)) != NULL)
  {
    if (*s == '[')
    {
      affy_unget_next_line(fp);
      break;
    }
    kv = split(s, '=');
    if (STREQ(kv[0], "Name"))
      cdf->array_type = strdup(kv[1]);
    else if (STREQ(kv[0], "Rows"))
      cdf->numrows = atoi(kv[1]);
    else if (STREQ(kv[0], "Cols"))
      cdf->numcols = atoi(kv[1]);
    else if (STREQ(kv[0], "NumQCUnits"))
      cdf->numqcunits = atoi(kv[1]);
    else if (STREQ(kv[0], "NumberOfUnits"))
      cdf->numprobesets = atoi(kv[1]);

  }

  /* At this point we know the size of the chip, and the number
     of probes.
   */
  cdf->cell_type = MALLOC(cdf->numrows * sizeof(BYTE *));
  cdf->xy_ref = MALLOC(cdf->numrows * sizeof(AFFY_PROBE **));
  for (i = 0; i < cdf->numrows; i++)
  {
    cdf->cell_type[i] = CALLOC(cdf->numcols, sizeof(BYTE));
    cdf->xy_ref[i] = CALLOC(cdf->numcols, sizeof(AFFY_PROBE *));
  }
  cdf->probeset = CALLOC(cdf->numprobesets, sizeof(AFFY_PROBESET));
  cdf->probe =
    CALLOC(cdf->numrows * cdf->numcols / 2.0, sizeof(AFFY_PROBE *));

  return;
}

/* A quality control section */
static void process_qc_section(FILE * fp, AFFY_CDFFILE *cdf, int qcnum)
{
  char *s, **kv;
  int num_qc_cells = 0, qc_counter;
  int x, y;

  /* Find number of cells, then cell header */
  while ((s = affy_get_next_line(fp)) != NULL)
  {
    kv = split(s, '=');
    if (STREQ(kv[0], "CellHeader"))
      break;
    if (STREQ(kv[0], "NumberCells"))
      num_qc_cells = atoi(kv[1]);
  }

  for (qc_counter = 0; qc_counter < num_qc_cells; qc_counter++)
  {
    s = affy_get_next_line(fp);
    kv = split(s, '=');
    sscanf(kv[1], "%d%d", &x, &y);
    cdf->cell_type[x][y] = AFFY_QC_LOCATION;
    cdf->xy_ref[x][y] = NULL;
  }

  return;
}

/* A probe section will be indexed by probe, and by x,y coordinates */
static void process_probe_section(FILE * fp, AFFY_CDFFILE *cdf, int ps)
{
  int status = AFFY_NORMAL_LOCATION;
  int numprobes = 0;
  char **kv, **f;
  char *s;
  int x, y, i, j;
  char pbase, tbase, atom;

  /* Find number of cells, then cell header */
  while ((s = affy_get_next_line(fp)) != NULL)
  {
    kv = split(s, '=');
    /* Name is important, also sets the location type */
    if (STREQ(kv[0], "Name"))
    {
      cdf->probeset[ps].name = strdup(kv[1]);
      if (!strncmp(kv[1], "AFFX", 4))
        status = AFFY_CONTROL_LOCATION;
    }
    else if (STREQ(kv[0], "NumAtoms"))
      numprobes = atoi(kv[1]);
    else if (STREQ(kv[0], "CellHeader"))
      break;
  }

  /* Allocate enough storage for all probes in this probe set */
  cdf->probeset[ps].probe = CALLOC(numprobes, sizeof(AFFY_PROBE));
  cdf->probeset[ps].numprobes = numprobes;
  cdf->probeset[ps].index = ps;

  /* Get each probe */
  for (i = 0; i < numprobes; i++)
  {
    for (j = 0; j < 2; j++)
    {
      s = affy_get_next_line(fp);
      kv = split(s, '=');

      f = split(kv[1], '\t');
      x = atoi(f[0]);
      y = atoi(f[1]);
      pbase = *f[7];
      tbase = *f[8];
      atom = *f[9];

      cdf->cell_type[x][y] = status;
      if (pbase != tbase)
      {
        cdf->probeset[ps].probe[i].pm.x = x;
        cdf->probeset[ps].probe[i].pm.y = y;
      }
      else
      {
        cdf->probeset[ps].probe[i].mm.x = x;
        cdf->probeset[ps].probe[i].mm.y = y;
      }
      cdf->xy_ref[x][y] = &(cdf->probeset[ps].probe[i]);
    }
    cdf->probeset[ps].probe[i].ps = &(cdf->probeset[ps]);

    /* Add probe to list of all probes */
    cdf->probe[cdf->numprobes] = &(cdf->probeset[ps].probe[i]);
    cdf->probe[cdf->numprobes]->index = cdf->numprobes;
    cdf->numprobes++;
  }

  return;
}
