
/**************************************************************************
 *
 * Filename: load_cel_file.c
 *
 * Purpose: Parse a CEL file and initialize an accompanying structure.
 *
 * Creation date: 
 *
 * Author: Steven Eschrich
 *
 *
 * Update History
 * --------------
 * 4/8/05: Imported/repaired from old libaffy (AMH)
 * 4/19/05: Now we use AFFY_CELL and collect all cell data (AMH)
 *
 **************************************************************************/

#include <affy.h>
#include <utils.h>

static void process_cel_section();
static void process_header_section();
static void process_intensity_section();
static void process_mask_section();
static void process_outlier_section();
static void process_modified_section();

static int pb;

void affy_load_text_cel_file(FILE *fp, AFFY_CELFILE *cf, int cf_pb)
{
  char *str;
 	
 	/* Allocate a big buffer for reading */	
  setvbuf(fp, NULL, _IOFBF, MAXBUF);		
 	
 	pb=cf_pb;
 	
  while ((str = affy_get_next_line(fp)) != NULL)
  {

    /* Process by section */
    if (STREQ(str, "[CEL]"))
      process_cel_section(fp, cf);
    else if (STREQ(str, "[HEADER]"))
      process_header_section(fp, cf);
    else if (STREQ(str, "[INTENSITY]"))
      process_intensity_section(fp, cf);
    else if (STREQ(str, "[MASKS]"))
      process_mask_section(fp, cf);
    else if (STREQ(str, "[OUTLIERS]"))
      process_outlier_section(fp, cf);
    else if (STREQ(str, "[MODIFIED]"))
      process_modified_section(fp, cf);
    /* This case should probably not occur */
    else
    {
      info(("    (Skipping the %s section.)\n", str));
      affy_skip_to_next_header(fp);
    }
  }

  return;
}

/*
  We have just read the [CEL] line. Process the remainder of the
  section until we get to eof or another []. 
*/
static void process_cel_section(FILE * fp, AFFY_CELFILE *cf)
{
  char *s;
  char **kv;

  /* Read lines until eof or a new section */
  while ((s = affy_get_next_line(fp)) != NULL)
  {
    if (*s == '[')
    {
      affy_unget_next_line(fp);
      return;
    }
    kv = split(s, '=');

    if (STREQ(kv[0], "Version"))
      info(("   CEL version:  %s\n", kv[1]));

  }

  return;
}

static void process_header_section(FILE * fp, AFFY_CELFILE *cf)
{
  char *s;
  char **kv;
  int i, n;

  /* Read lines until eof or a new section */
  while ((s = affy_get_next_line(fp)) != NULL)
  {
    if (*s == '[')
    {
      affy_unget_next_line(fp);
      break;
    }
    kv = split(s, '=');
    if (STREQ(kv[0], "Cols"))
      cf->numcols = atoi(kv[1]);
    else if (STREQ(kv[0], "Rows"))
      cf->numrows = atoi(kv[1]);
  }

  /* Once we know the parameters, we allocate storage */
  n = numbytes(cf->numcols);
  cf->data = MALLOC(cf->numrows * sizeof(AFFY_CELL *));
  cf->data[0] = CALLOC(cf->numcols * cf->numrows, sizeof(AFFY_CELL));
  cf->mask = MALLOC(cf->numrows * sizeof(affy_uint8 *));
  cf->mask[0] = CALLOC(n * cf->numrows, sizeof(affy_uint8));
  cf->outlier = MALLOC(cf->numrows * sizeof(affy_uint8 *));
  cf->outlier[0] = CALLOC(n * cf->numrows, sizeof(affy_uint8));

  for (i = 1; i < cf->numrows; i++)
  {
    cf->data[i] = cf->data[0] + (i * cf->numcols);
    cf->mask[i] = cf->mask[0] + (i * n);
    cf->outlier[i] = cf->outlier[0] + (i * n);
  }

  info(("   Dim: %dx%d\n", cf->numrows, cf->numcols));

  return;
}

static void process_intensity_section(FILE * fp, AFFY_CELFILE *cf)
{
  char *s;
  char **kv;
  Boolean read_cellheader = False;
  int x, y, npixels;
  double val, stdv;
  int num_read = 0, progress, incr;

  incr = (cf->numrows * cf->numcols) / 100;
  progress = incr;

  pb_update(pb, 1, "Loading intensities");
  /* Read lines until eof or a new section */
  while ((s = affy_get_next_line(fp)) != NULL)
  {
    if (*s == '[')
    {
      affy_unget_next_line(fp);
      break;
    }

    /* Only do streq's in case we haven't read the header */
    if (!read_cellheader)
    {
      kv = split(s, '=');
      if (STREQ(kv[0], "NumberCells"));
      else if (STREQ(kv[0], "CellHeader"))
        read_cellheader = True;
    }
    else
    {
      /* Otherwise, this is a line of x,y coordinates and mean intensity */
      if (num_read == progress)
      {
        progress += incr;
        pb_update(pb, progress / incr, "");
      }
      num_read++;
      sscanf(s, "%d %d %lf %lf %d", &x, &y, &val, &stdv, &npixels);
      cf->data[x][y].value = val;
      cf->data[x][y].stddev = stdv;
      cf->data[x][y].numpixels = npixels;
      bset(cf->mask[x], y, 1);
      bset(cf->outlier[x], y, 1);
    }
  }

  return;
}

static void process_mask_section(FILE * fp, AFFY_CELFILE *cf)
{
  char *s;
  char **kv;
  Boolean read_maskheader = False;
  int num_masks = 0;
  int x, y;

  pb_update(pb, 1, "Processing Masks");

  /* Read lines until eof or a new section */
  while ((s = affy_get_next_line(fp)) != NULL)
  {
    if (*s == '[')
    {
      affy_unget_next_line(fp);
      break;
    }
    /* Only do streq's in case we haven't read the header */
    if (!read_maskheader)
    {
      kv = split(s, '=');
      if (STREQ(kv[0], "NumberCells"))
      {
        cf->nummasks = atoi(kv[1]);
        info(("(%d masks)...", cf->nummasks));
      }
      else if (STREQ(kv[0], "CellHeader"))
        read_maskheader = True;
    }
    else
    {
      /* Otherwise, this is a line of x,y coordinates and mean intensity */
      sscanf(s, "%d %d", &x, &y);
      bset(cf->mask[x], y, 0);
      num_masks++;
    }
  }

  if (num_masks != cf->nummasks)
    warn(("Mismatch on number of masks: %d actual, %d expected\n",
          num_masks, cf->nummasks));

  return;
}

static void process_outlier_section(FILE * fp, AFFY_CELFILE *cf)
{
  register char *s;
  char **kv;
  Boolean read_outlierheader = False;
  int num_outliers = 0;
  int x, y;

  pb_update(pb, 1, "Processing Outliers");
  /* Read lines until eof or a new section */
  while ((s = affy_get_next_line(fp)) != NULL)
  {
    if (*s == '[')
    {
      affy_unget_next_line(fp);
      break;
    }
    /* Only do streq's in case we haven't read the header */
    if (!read_outlierheader)
    {
      kv = split(s, '=');
      if (STREQ(kv[0], "NumberCells"))
      {
        cf->numoutliers = atoi(kv[1]);
        info(("(%d outliers)...", cf->numoutliers));
      }
      else if (STREQ(kv[0], "CellHeader"))
        read_outlierheader = True;
    }
    else
    {
      /* Otherwise, this is a line of x,y coordinates and mean intensity */
      sscanf(s, "%d %d", &x, &y);
      bset(cf->outlier[x], y, 0);
      num_outliers++;
    }
  }

  if (num_outliers != cf->numoutliers)
    warn(("Mismatch on number of outliers: %d actual, %d expected\n",
          num_outliers, cf->numoutliers));

  return;
}

/* This section is supposed to be outdated post MAS 4, so
   it should be empty and ignored.
*/
static void process_modified_section(FILE * fp, AFFY_CELFILE *cf)
{

  affy_skip_to_next_header(fp);
  return;
}
