#ifndef __UTILS_H_
#define __UTILS_H_

/* 
 *  utils.h
 *
 *  Adapted from the original libutils, written by:
 *
 *       Steven Eschrich
 *	 eschrich@csee.usf.edu
 *	 
 *  ANSI-ified by Andrew M. Hoerter.  Where possible I have kept to the
 *  original libutils interfaces, but in some cases changes have been
 *  made to make the code standards-conforming.  Any resulting bugs are
 *  mine alone.
 *
 */

#include "endian_config.h"

#define __UTILS_VERSION  "$Id: utils.h -- libaffy variant (AMH)"

/* System defines needed somewhere below */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>

#include "utils_types.h"

/********************************************************
 * Compiler-specific knobs.
 * 
 ********************************************************/
#if defined(__GNUC__) && !defined(__STRICT_ANSI__)
/* GCC-isms. */
#define INLINE inline
#define UNUSED __attribute ((unused))

#elif (__STDC_VERSION__ >= 199901L)
/* C99-isms. */
#define INLINE inline
#define UNUSED

#else
/* The rest of the world is C89. */
#define INLINE
#define UNUSED
#endif

/********************************************************
 * Constants defined within libutils.
 * 
 ********************************************************/

/* The buffer size generally to create statically, a large number */
#define MAXBUF        (10*1024)

/* If arrays grow, by what increment */
#define REALLOC_INCR  (1024)

/********************************************************
 * Variables defined within the library.
 * 
 ********************************************************/

/* An indicator of the endian-ness of the system */
extern int endian;

/* Constant bit-masks for simplicity. */
extern BYTE setbits[8];

/*****************************************************************************
   Macros
 
 	 Various useful macros (die, warn, info, etc). These macros ultimately
 	 call implementations after doing some massaging.

*****************************************************************************/

void set_locus(const char *fname, unsigned int line);

/* Sensible string comparison */
#define STREQ(a,b)   ( !strcasecmp(a,b) )

/* Many logging utilities are defined separately */
#include "util_log.h"

#define MALLOC(size)     (set_locus(__FILE__, __LINE__), MALLOC_impl(size))
#define CALLOC(n, size)  (set_locus(__FILE__, __LINE__), CALLOC_impl(n, size))
#define REALLOC(q, n)    (set_locus(__FILE__, __LINE__), REALLOC_impl(q, n))
#define FOPEN(fname, mode) (set_locus(__FILE__, __LINE__), \
                            FOPEN_impl(fname, mode))

/* These simpler min/max macros are NOT side-effect safe! */
#define min(X, Y)                     \
        ((X < Y) ? X : Y)

#define max(X, Y)                     \
        ((X > Y) ? X : Y )

/* A perlism - chomp the last character off of the string */
#define chomp(q) \
      do { \
         register char *__p=q; \
         register int __plen=strlen(__p); \
         if ( __p != 0 && __plen > 0 ) \
	   __p[__plen-1]=0; \
      } while(0)

/* A perlism, push a new element on the given array */
#define push(array, element, array_count, array_type) \
      do { \
         register int __tmp_count=array_count; \
         set_locus(__FILE__, __LINE__);         \
         if ( __tmp_count % REALLOC_INCR == 0 ) \
             array=(array_type *)realloc(array, \
                                            (__tmp_count+REALLOC_INCR)* \
                                              sizeof(array_type)); \
         if ( ! array ) { \
           die_impl("Memory allocation failure");\
         }\
         array[__tmp_count]=element; \
      } while (0)

/* A perlism, pop the last element off of the array */
#define pop(array, array_count, array_type) \
        ({ \
          register int __tmp_count=array_count; \
          array_type __retval=0; \
          if ( __tmp_count-1 >= 0 ) \
             __retval=array[__tmp_count-1]; \
          __retval; \
        })

/*****************************************************************************
  SECTION2: Inline functions

  If you have GCC or a C99 compiler, these ought to be real inlines;
  otherwise they will be small static functions which a smart compiler
  will hopefully inline anyway.  If all else fails, oh well.
*****************************************************************************/

/* numbytes(n) - Given n bits, return number of bytes needed/used */
static INLINE int numbytes(int n)
{
  register unsigned int i = n;
  register unsigned int d = i >> 3;

  if ((i & 0x7) > 0)
    d++;

  return d;
} UNUSED

/* bget(v,index) - Return the index'th bit in the bitstring v */
static INLINE int bget(BYTE * v, int index)
{
  register unsigned int i = index;
  register unsigned int o = i & 0x7;
  register unsigned int d = i >> 3;

  return ((v[d] & setbits[o]) > 0);
} UNUSED

/* bset(v,index,value) - Set the index'th bit in the bitstring v to value */
static INLINE void bset(BYTE * v, int index, int value)
{
  register unsigned int i = index;
  register unsigned int o = i & 0x07;
  register unsigned int d = i >> 3;

  if ((value) == 0)
    v[d] &= (BYTE) ~ setbits[o];
  else
    v[d] |= setbits[o];

  return;
} UNUSED

/* Remove comments from line, given comment char c */
static INLINE int strip_comments(char *p, char c)
{
  if (!p)
    return -1;
  for (; *p != 0; p++)
  {
    if (*p == c)
    {
      *p = 0;
      break;
    }
  }
  return 0;
} UNUSED

/* Trim leading spaces (whitespace) */
static INLINE char *ltrim(char *start)
{
  if (!start)
    return start;
  for (; *start != 0 && isspace((int)(*start)); start++);
  return start;
} UNUSED

/* Seek from the end of p backwards to the first non-blank character.
   This is a bit difficult since array bounds must be kept in mind. */
static INLINE char *rtrim(char *start, char *end)
{
  if (!start)
    return 0;
  /* Seek back to beginning or first non-blank */
  for (; end > start && (*end == 0 || isspace((int)(*end))); end--);
  /* Are we at the beginning? If so, we might either have a blank
     or a non-blank. */
  if (end == start)
  {
    if (isspace((int)(*end)))
      *end = 0;                 /* \n or the like */
    else if (*end != 0)
      *(end + 1) = 0;           /* non-empty string, zero next char */
  }
  else
    *(end + 1) = 0;
  return end;
} UNUSED

/*
  endsWith(). Like the java sort of version, does it end with the
  given string. See also startsWith(). Does case insensitive comparison.
*/
static INLINE int endsWith(char *str, char *sub)
{
  int N = strlen(str);
  int last = strlen(sub);

  if (last > N)
    return 0;

  return (!strcasecmp(str + N - last, sub));
} UNUSED

/* MALLOC space, exiting the program on failure */
static INLINE void *MALLOC_impl(size_t size)
{
  register void *result;

  if ((result = malloc(size)) == NULL)
  {
    die_impl("error allocating memory");
    return (NULL);              /* To silence compiler diagnostics. */
  }
  else
    return (result);
} UNUSED

/* CALLOC space, exiting the program on failure */
static INLINE void *CALLOC_impl(size_t n, size_t size)
{
  register void *result;

  if ((result = calloc(n, size)) == NULL)
  {
    die_impl("error allocating memory");
    return (NULL);              /* To silence compiler diagnostics. */
  }
  else
    return (result);
} UNUSED

/* REALLOC space, exiting the program on failure */
static INLINE void *REALLOC_impl(void *q, size_t n)
{
  register void *result;

  if ((result = realloc(q, n)) == NULL)
  {
    die_impl("error re-allocating memory");
    return (NULL);              /* To silence compiler diagnostics. */
  }
  else
    return (result);
} UNUSED

/* Open a file, exiting program on failure. */
static INLINE void *FOPEN_impl(const char *filename, const char *mode)
{
  FILE *result;

  if ((result = fopen(filename, mode)) == NULL)
  {
    die_impl("Error opening %s (mode %s): %s", filename, mode,
             strerror(errno));
    return (NULL);              /* To silence compiler diagnostics. */
  }
  else
    return (result);
} UNUSED

/*****************************************************************************
   SECTION3: Regular functions

   Here are functions that are normally implemented within libutils.a for
   general use. There is no need to inline, etc.
*****************************************************************************/

#ifdef __cplusplus
extern "C"
{
#endif

  char *trim();
  char **split();

/* Compare two doubles (for sorting) */
  int dcompare(const void *p1, const void *p2);
  int icompare(const void *p1, const void *p2);
  int fcompare(const void *p1, const void *p2);

/* Matrix allocation */
  double **create_matrix(int rows, int columns);
  void free_matrix(double **m);

  int swap_int(unsigned int i);
  short swap_short(unsigned short i);

#ifdef __cplusplus
};
#endif

#endif
