
/**************************************************************************
 *
 * Filename: mas5_background_correction.c
 *
 * Purpose: 
 *
 * Creation date: 
 *
 * Author: Steven Eschrich
 *
 *
 * Update History
 * --------------
 * 4/14/05: Imported/repaired from old libaffy (AMH)
 * 4/19/05: AFFY_CELFILE now uses AFFY_CELL's (AMH)
 *
 **************************************************************************/

#include <affy_mas5.h>


#define START_X 0
#define START_Y 1
#define LENGTH_X 2
#define LENGTH_Y 3

static int K;                   /* The number of rectangular zones on the chip */
static int smooth;
static double NoiseFrac;

AFFY_POINT *center;
double *bZ;
double *nZ;
int dim, default_grid_y_length, default_grid_x_length;

int find_centers();
int estimate_zone_background();
int calculate_background();
int background();
double w_k();

/*
  This operation is detailed in the Affy white papers. It takes
  a CELFILE object, background corrects it and returns.
*/
int mas5_background_correction(AFFY_CHIPSET *c, MAS5_FLAGS *f)
{
  int n;

  if (f == NULL)
    f = mas5_get_defaults();
  K = f->K;
  smooth = f->smooth;
  NoiseFrac = f->NoiseFrac;

  info(("  Background correction using Affymetrix method.\n"));

  /* Compute these dynamically since we don't know K a priori */
  dim = sqrt(K);
  default_grid_x_length = c->numcols / dim;
  default_grid_y_length = c->numrows / dim;

  /* Allocate space for these variables now */
  center = CALLOC(K, sizeof(AFFY_POINT));
  bZ = CALLOC(K, sizeof(double));
  nZ = CALLOC(K, sizeof(double));

  /*
     Three steps: find grid centers, estimate zone background, then
     finally compute the background intensities. The background is not
     only calculated but intensities are adjusted relative to this
     background.
   */
  info(("   Finding centers:..."));
  find_centers(c->numrows, c->numcols);
  info(("done.\n"));
  info(("   Estimating zone background and calculating background correction:"));
  for (n = 0; n < c->numchips; n++)
  {
    info(("."));
    estimate_zone_background(c->chip[n]);
    calculate_background(c->chip[n]);
  }
  info(("done.\n"));

  free(center);
  free(bZ);
  free(nZ);

  return 0;
}

/* A simple calculation of grid centers, based on a square grid of size K */
int find_centers(int rows, int cols)
{
  int running_x_offset, running_y_offset;
  int k;
  int lengthx=zone_information(k,LENGTH_X);
  int lengthy=zone_information(k,LENGTH_Y);
  int midy = lengthy / 2;
  int midx = lengthx / 2;

  /* Fill in the center points based on K */
  running_x_offset = running_y_offset = 0;
  for (k = 0; k < K; k++)
  {
    /* Increment center if we wrap around a row */
    if (running_x_offset == cols)
    {
      running_x_offset = 0;
      running_y_offset += lengthy;
    }
    /* Calculation is halfway into next region */
    center[k].x = running_x_offset + midx;
    running_x_offset += lengthx;

    center[k].y = running_y_offset + midy;
  }

  return 0;
}


/*
   A zone is bit weird but here is the deal, thanks to perls of
   wisdom gleaned from GNU Affymetrix code:
       A zone cannot have split PM/MM pairs. Probe pairs appear to be
      setup as ODDV then EVEN, so that the ODD is always the
      smaller number.
	 
      What this means to zones is that the fixed length grid is not
      fixed, since it must be adjusted to include an extra line at 
      the beginning (which is QC anyway) and must be an extra pair
      short at the end (the extra one from the top plus the QC row 
      at the bottom).
     

   Given a particular zone
*/
int zone_information(int k, int type) {
 
   int tmp;
   
   switch (type) {
      case START_X:
         /* The X coordinates are constant, so this is a simple calculation */
         return (k%dim) * default_grid_x_length;

      case START_Y:
      	 /* If k is part of the first row, start at 0. */
	 if ( k / dim == 0 )
	    return 0;
	 /* Otherwise, the start is around default but must be odd */
	 tmp=(k/dim) * default_grid_y_length;
	 if ( tmp % 2 == 0 ) tmp++;
	 return tmp; 
	 
      case LENGTH_X:
      	 return default_grid_x_length;

      case LENGTH_Y:
      	 /* If k is not first or last row, use default */
	 tmp=k/dim;

      	 /* First row, if length of grid is even, add one */
	 if ( tmp == 0 ) {
      	    if ( default_grid_y_length % 2 == 0 )
	       return default_grid_y_length+1;
	    else
	       return default_grid_y_length;
	 }
	 /* Last row, if length of grid is even, subtract one */
	 if ( tmp == dim  - 1 ) {
      	    if ( default_grid_y_length % 2 == 0 )
	       return default_grid_y_length-1;
	    else
	       return default_grid_y_length;
      	 }
	 return default_grid_y_length;
   }

   return -1; 
   
}

/* This should calculate the lower 2% cutoff and stddev */
int estimate_zone_background(AFFY_CHIP *chip)
{
  int k, i, x, y;
  int num_in_bg, num_bgvals, total_vals;
  int startx, starty;
  double *bgvals;
  AFFY_CELFILE *cf = chip->cel;

  /* What is 2% of a zone. This is the upper bound. */
  num_in_bg = 0.02 * (default_grid_x_length * (default_grid_y_length+1));
  bgvals = CALLOC(num_in_bg, sizeof(double));

  /* For each zone, copy the values to a separate array to work on */
  for (k = 0; k < K; k++)
  {
    int starty=zone_information(k,START_Y);
    int startx=zone_information(k,START_X);
    int lengthy=zone_information(k,LENGTH_Y);
    int lengthx=zone_information(k,LENGTH_X);

    num_bgvals = 0;
    total_vals = 0;


    /* IMPORTANT: Zero out this area for each zone */ 
    for (i=0; i < num_in_bg; i++) 
       bgvals[i]=-1;

    /* Accumulate lower 2% of region */
     for (y = 0; y < lengthy; y++)
    {
      for (x = 0; x < lengthx; x++)
      {
        int ry = starty + y;    /* Real y offset */
        int rx = startx + x;    /* Real x offset */

	/* Skip over masked cells and undefined cells and QC cells */
        if (affy_ismasked(chip, rx, ry) || affy_isundefined(chip, rx, ry) ||
	    affy_isqc(chip,rx,ry) )
          continue;

        /* Insert into bgvals list, in sorted order */
        if (num_bgvals < num_in_bg)
        {
          /* Insert into list */
          for (i = num_bgvals;
               i > 0 && bgvals[i - 1] > cf->data[rx][ry].value; i--)
            bgvals[i] = bgvals[i - 1];
          bgvals[i] = cf->data[rx][ry].value;
          num_bgvals++;
        }
        else if (bgvals[num_in_bg - 1] < cf->data[rx][ry].value)
        {
          /* Do nothing */
        }
        else
        {
          /* Insert into list */
          for (i = num_in_bg - 1;
               i > 0 && bgvals[i - 1] > cf->data[rx][ry].value; i--)
            bgvals[i] = bgvals[i - 1];
          bgvals[i] = cf->data[rx][ry].value;
        }
      }
    }

     /* mean of the lower 2% */
    num_bgvals=0.02*default_grid_x_length*default_grid_y_length;
    bZ[k] = 0;
    for (i = 0; i < num_bgvals; i++)
      bZ[k] += bgvals[i];
    bZ[k] /= num_bgvals;


    /* Standard deviation */
    nZ[k] = 0;
    for (i = 0; i < num_bgvals; i++)
      nZ[k] += (bgvals[i] - bZ[k]) * (bgvals[i] - bZ[k]);
    nZ[k] = sqrt(nZ[k] / (num_bgvals - 1));

  }
#ifdef DEBUG
  /* Output the various statistics that come from this background */
  output_statistics();
#endif
  
  free(bgvals);

  return 0;
}

int output_statistics() {
  double totalBG=0;
  double totalN=0;
  int k;
  
  for (k=0; k < K; k++) {
     totalBG+=bZ[k];
     totalN+=nZ[k];
  }
   totalBG/=K;
   totalN/=K;
   for (k=0; k < K; k++) 
     info(("Background[%d]=%f\n",k,bZ[k]));
   info(("Average BG is %f\n",totalBG));
   for (k=0; k < K; k++) 
     info(("Noise[%d]=%f\n",k,nZ[k]));
   info(("Average N is %f\n",totalN));
}


/* A Byzantine calculation involving many substeps */
int calculate_background(AFFY_CHIP *chip)
{
  double b, n;
  int x, y;
  double I_prime;
  AFFY_CELFILE *cf = chip->cel;
  int pinterval = cf->numrows / 10;
  int progress = pinterval;

  for (y = 0; y < cf->numrows; y++)
  {
    if (y == progress)
    {
      progress += pinterval;
    }
    for (x = 0; x < cf->numcols; x++)
    {

      /* Calculate both the b and n values */
      background(x, y, &b, &n);
      I_prime = max(cf->data[x][y].value, 0.5);
      cf->data[x][y].value = max(I_prime - b, NoiseFrac * n);
    }
  }
  return 0;
}

/* This is the b(x,y) calculation */
int background(int x, int y, double *b, double *n)
{
  double denom = 0;
  double n_n = 0, b_n = 0;
  double cur_weight;
  int k;

  for (k = 0; k < K; k++)
  {
    cur_weight = w_k(x, y, k);
    denom += cur_weight;
    b_n += cur_weight * bZ[k];
    n_n += cur_weight * nZ[k];
  }
  *b = b_n / denom;
  *n = n_n / denom;

  return 0;
}

/* This is the w_k(row,col) calculation */
double w_k(int x, int y, int k)
{
  double d;

  d =
    ((x - center[k].x) * (x - center[k].x)) +
    ((y - center[k].y) * (y - center[k].y));
  return 1.0 / (d + smooth);
}
