
/**************************************************************************
 *
 * Filename: mas5_signal.c
 *
 * Purpose: 
 *
 * Creation date: 
 *
 * Author: Steven Eschrich
 *
 *
 * Update History
 * --------------
 * 4/14/05: Imported/repaired from old libaffy (AMH)
 * 4/19/05: AFFY_CELFILE uses AFFY_CELL's now (AMH)
 *
 **************************************************************************/

#include <affy_mas5.h>

/* Weird constants used by affy */
double delta;
double contrast_tau;
double scale_tau;

double LOG2 = -1;

/* Private routines */
static int double_sort_compare();
static double median();
static double tukey_biweight();
static double calculate_specific_background();
static double calculate_probeset_signal();

int mas5_signal(AFFY_CHIPSET *c, MAS5_FLAGS *f)
{
  int i, n;
  int progress, incr;
  int num_probesets;

  if (c == 0 || c->numchips == 0 || c->cdf == 0)
    return -1;
  if (f == NULL)
    f = mas5_get_defaults();
  contrast_tau = f->contrast_tau;
  scale_tau = f->scale_tau;
  delta = f->delta;

  LOG2 = log(2);
  num_probesets = c->cdf->numprobesets;

  for (n = 0; n < c->numchips; n++)
  {
    progress = incr = c->cdf->numprobesets / 10;
    info(("  Calculating signal for chip using Affymetrix method:"));

    c->chip[n]->probe_set = CALLOC(num_probesets, sizeof(double));
    c->chip[n]->numprobesets = num_probesets;
    for (i = 0; i < num_probesets; i++)
    {
      if (i == progress)
      {
        info(("."));
        progress += incr;
      }
      c->chip[n]->probe_set[i] = calculate_probeset_signal(c->chip[n], i);
    }
    info(("done.\n"));
  }

  return 0;
}

static double calculate_probeset_signal(AFFY_CHIP *c, int probeset_num)
{
  AFFY_PROBESET *p;
  AFFY_CELL **data;
  double *pm, *mm, *pv;
  double signal, signal_log_value;
  int n, i, j;
  double SB;
  int x, y;
  double im;

  /* Some shortcuts */
  p = &(c->cdf->probeset[probeset_num]);
  n = p->numprobes;
  data = c->cel->data;

  /* Assign pm/mm values for this chip (based on existing layout) */
  pm = CALLOC(n, sizeof(double));
  mm = CALLOC(n, sizeof(double));
  pv = CALLOC(n, sizeof(double));

  /* Calculate the SB. NOTE: For some reason, it is calculated
     on all PM/MM probe pairs, even if they are masked. So that
     must be done in advance here. */
  for (i = 0; i < n; i++)
  {
    x = p->probe[i].pm.x;
    y = p->probe[i].pm.y;
    pm[i] = data[x][y].value;
    x = p->probe[i].mm.x;
    y = p->probe[i].mm.y;
    mm[i] = data[x][y].value;
  }
  SB = calculate_specific_background(pm, mm, n);

  /*
     Assign pm/mm values based on probes. Caveat: masked probes (either 
     in PM or MM) are ignored in the computation.
   */
  for (i = 0, j = 0; i < n; i++)
  {
    x = p->probe[i].pm.x;
    y = p->probe[i].pm.y;
    if (affy_ismasked(c, x, y))
      continue;
    pm[j] = data[x][y].value;

    x = p->probe[i].mm.x;
    y = p->probe[i].mm.y;
    if (affy_ismasked(c, x, y))
      continue;
    mm[j] = data[x][y].value;
    j++;
  }
  /* In case the total number of probes is less than n */
  if (n > j)
  {
#ifdef DEBUG
    info(("%s:Adjusting for %d probes instead of %d\n", p->probe_set_id, j,
          n));
#endif
    n = j;
  }

  /* Calculate Probe Values (PV) */
  for (j = 0; j < n; j++)
  {

    /* Easiest case, mm < pm */
    if (pm[j] > mm[j])
    {
      im = mm[j];
    }
    else
    {
      /* Based on SB, calculate IM */
      if (SB - contrast_tau > 0)
        im = pm[j] / pow(2, SB);
      else
        im =
          pm[j] / pow(2,
                      (contrast_tau /
                       (1 + ((contrast_tau - SB) / scale_tau))));
    }

    /* Once im is determined, calculate pv */
    pv[j] = log(max(pm[j] - im, delta)) / LOG2;
  }

  /* Signal log value is here */
  signal_log_value = tukey_biweight(pv, n);

  /* Then take the antilog and we're done */
  signal = pow(2, signal_log_value);

  free(pm);
  free(mm);
  free(pv);

  return signal;
}

/*----------------------------------------------------------------------
   The specific background is the Tukey Biweight of log differences
 -----------------------------------------------------------------------*/
static double calculate_specific_background(double *pm, double *mm, int n)
{
  double *d;
  double r;
  int j;

  d = MALLOC(n * sizeof(double));

  /* Difference of log values */
  for (j = 0; j < n; j++)
    d[j] = (log(pm[j]) / LOG2 - log(mm[j]) / LOG2);

  /* Compute this biweight */
  r = tukey_biweight(d, n);

  free(d);

  return r;
}

/*--------------------------------------------------------------------
  Calculate Tukey's Biweighted Average, per the Affy docs.
  --------------------------------------------------------------------*/
static int c = 5;
static double epsilon = 0.0001;

static double tukey_biweight(double *x, int n)
{
  double *u;
  int i;
  double M, S;
  double Tbi_num = 0, Tbi_denom = 0;
  double *diffs;

  u = MALLOC(n * sizeof(double));

  /* Calculate median */
  M = median(x, n);

  /* Calculate S, median of absolute differences from M */
  diffs = MALLOC(n * sizeof(double));
  for (i = 0; i < n; i++)
    diffs[i] = fabs(x[i] - M);
  S = median(diffs, n);
  free(diffs);

  /* Calculate distance measure */
  for (i = 0; i < n; i++)
    u[i] = (x[i] - M) / (c * S + epsilon);

  /* Finally, calculate the result */
  for (i = 0; i < n; i++)
  {
    double usquared, w;

    /* Function w(u) is 0 for all |u|>0 */
    if (fabs(u[i]) > 1)
      continue;

    usquared = u[i] * u[i];
    w = (1.0 - usquared) * (1.0 - usquared);
    Tbi_num += w * x[i];
    Tbi_denom += w;
  }

  free(u);

  return Tbi_num / Tbi_denom;
}

/*
  Calculate the median of n numbers (x[0]..x[n-1]) without touching
  the x array.
*/
static double median(double *x, int n)
{
  double *d;
  int i;
  double M;

  /* Copy the array */
  d = MALLOC(n * sizeof(double));
  for (i = 0; i < n; i++)
    d[i] = x[i];

  /* Sort in increasing order */
  qsort(d, n, sizeof(double), double_sort_compare);
  /* Median is middle value, or mean of two middle values */
  if (n % 2 == 1)
    M = d[n / 2];
  else
    M = (d[(n / 2) - 1] + d[n / 2]) / 2.0;

  free(d);
  return M;
}

static int double_sort_compare(void *p1, void *p2)
{
  double d1 = *((double *)p1);
  double d2 = *((double *)p2);

  if (d1 < d2)
    return -1;
  if (d1 > d2)
    return 1;
  if (d1 == d2)
    return 0;

  return -1;

}
