
/**************************************************************************
 *
 * Filename: median_polish.c
 *
 * Purpose: 
 *
 * Creation date: 
 *
 * Author: Steven Eschrich
 *
 *
 * Update History
 * --------------
 * 4/14/05: Imported/repaired from old libaffy (AMH)
 *
 **************************************************************************/

/* This function was adapted from the bioconductor package */

/*****************************************************************************
 
  void median_polish(double *data, int rows, int cols, int *cur_rows, 
		     double *results, int nprobes)

  double *data - a data matrix of dimension rows by cols (the entire PM matrix)
  int rows, cols - rows and columns dimensions of matrix
  int cur_rows - vector of length nprobes containg row indicies of *data 
                 matrix which apply for a particular probeset
  double *results - a vector of length cols already allocated. on output 
		   contains expression values
  int nprobes - number of probes in current probeset.

  a function to do median polish.

 *****************************************************************************/

#include <affy_rma.h>

static int max_iterations = 10;

void get_column_median();

/* Add xdelta to x, for n elements */
static INLINE void vector_add(double *x, double *xdelta, int n)
{
  int i;

  for (i = 0; i < n; i++)
    x[i] += xdelta[i];
  return;
}

static INLINE double sum_abs(double **z, int rows, int cols)
{

  int i, j;
  double sum = 0.0;

  for (i = 0; i < rows; i++)
    for (j = 0; j < cols; j++)
      sum += fabs(z[i][j]);

  return sum;
}

static INLINE void subtract_by_row(double **z, double *rdelta, int rows,
                                   int cols)
{

  int i, j;

  for (i = 0; i < rows; i++)
    for (j = 0; j < cols; j++)
      z[i][j] -= rdelta[i];
  return;
}

static INLINE void subtract_by_col(double **z, double *cdelta, int rows,
                                   int cols)
{

  int i, j;

  for (j = 0; j < cols; j++)
  {
    for (i = 0; i < rows; i++)
    {
      z[i][j] -= cdelta[j];
    }
  }

}

/*
  Do a median polish on the chipset c, using probe set # ps. The results
  are stored in the corresponding probe_set array for each chip.
*/
void median_polish(double **z, int numprobes, int numchips,
                   double *results, double *affinities)
{
  double *rdelta, *r, *cdelta, *c;
  double eps = 0.01;
  double oldsum = 0.0, newsum = 0.0;
  double t = 0.0;
  double delta;
  int i, j, it;

  rdelta = CALLOC(numprobes, sizeof(double));
  cdelta = CALLOC(numchips, sizeof(double));
  r = CALLOC(numprobes, sizeof(double));
  c = CALLOC(numchips, sizeof(double));

  /* Do the actual median polish */
  for (it = 0; it < max_iterations; it++)
  {

    /* Rows first */
    affy_get_row_median(z, rdelta, numprobes, numchips);
    subtract_by_row(z, rdelta, numprobes, numchips);
    vector_add(r, rdelta, numprobes);
    delta = affy_median_save(c, numchips);
    for (j = 0; j < numchips; j++)
      c[j] -= delta;
    t += delta;

    /* Then columns */
    affy_get_column_median(z, cdelta, numprobes, numchips);
    subtract_by_col(z, cdelta, numprobes, numchips);
    vector_add(c, cdelta, numchips);
    delta = affy_median_save(r, numprobes);
    for (i = 0; i < numprobes; i++)
      r[i] -= delta;
    t += delta;

    /* If changes are small, then quit */
    newsum = sum_abs(z, numprobes, numchips);
    if (newsum == 0.0 || fabs(1.0 - oldsum / newsum) < eps)
      break;
    oldsum = newsum;
  }

  /* The result is the expression value */
  for (j = 0; j < numchips; j++)
    results[j] = t + c[j];
  if (affinities != NULL)
  {
    for (j = 0; j < numprobes; j++)
      affinities[j] = t + r[j];
  }

  free(rdelta);
  free(cdelta);
  free(r);
  free(c);

  return;
}
