
/**********************************************************
 **
 ** file: quantile_normalization.c
 **
 ** aim: A c implementation of the quantile normalization method 
 **
 ** Copyright (C) 2002-2003    Ben Bolstad
 **
 ** written by: B. M. Bolstad  <bolstad@stat.berkeley.edu>
 **
 ** written: Feb 2, 2002
 ** last modified: Apr 19, 2002
 ** 
 ** This c code implements the quantile normalization method
 ** for normalizing high density oligonucleotide data as discussed
 ** in
 **
 ** Bolstad, B. M., Irizarry R. A., Astrand, M, and Speed, T. P. (2003)(2003) 
 ** A Comparison of Normalization Methods for High 
 ** Density Oligonucleotide Array Data Based on Bias and Variance.
 ** Bioinformatics 19,2,pp 185-193
 **
 ** History
 ** Feb 2, 2002 - Intial c code version from original R code
 ** Apr 19, 2002 - Update to deal more correctly with ties (equal rank)
 ** Jan 2, 2003 - Documentation/Commenting updates reformating
 ** Feb 17, 2003 - add in a free(datvec) to qnorm(). clean up freeing of dimat
 ** Feb 25, 2003 - try to reduce or eliminate compiler warnings (with gcc -Wall)
 ** Feb 28, 2003 - update reference to normalization paper in comments
 ** Mar 25, 2003 - ability to use median, rather than mean in so called "robust" method
 **
 ***********************************************************/

#include <affy_rma.h>

extern int rma_pb;

typedef struct
{
  double data;
  int rank;
} dataitem;

int qnorm_compare(const void *p1, const void *p2);
void rank_order(double *rank, dataitem * x, int n);

void quantile_normalization_chip(AFFY_CHIPSET *c, int chipnum,
                                 double *mean)
{
  int i;
  double *rank;
  dataitem *vals;
  int num_probes;

  assert(c != NULL);
  assert(c->cdf != NULL);
  assert(mean != NULL);

  num_probes = c->cdf->numprobes;

  pb_update(rma_pb, 4, "Quantile normalization");

  /* Allocate storage */
  vals = CALLOC(num_probes, sizeof(dataitem));

  /* Load in pm for each probe */
  for (i = 0; i < num_probes; i++)
  {
    vals[i].data = c->chip[chipnum]->pm[i];
    vals[i].rank = i;
  }

  qsort(vals, num_probes, sizeof(dataitem), qnorm_compare);

  /* Step two: accumulate mean value at a given rank */
  for (i = 0; i < num_probes; i++)
  {
    mean[i] += vals[i].data;
#if 1
    if (vals[i].data > 100000) {
      info(("val = %f\n", vals[i].data));
      exit(1);
    }
    if(isnan(mean[i])) {
      info(("Mean[%d] is NAN\n", i));
      exit(1);
    }
#endif
  }

  rank = CALLOC(num_probes, sizeof(double));
  /* Rank order the intensities on this chip */
  rank_order(rank, vals, num_probes);
  for (i = 0; i < num_probes; i++)
  {
    int r = vals[i].rank;

    c->chip[chipnum]->pm[r] = floor(rank[i]) - 1;
  }
  free(rank);
  free(vals);
  pb_update(rma_pb, 5, "");
  return;
}

/**
 * Given rankings already computed per chip and an overall mean
 * profile, we can normalize each chip.
 */
void quantile_normalization_chipset(AFFY_CHIPSET *c, double *mean)
{
  int i, j;
  int numprobes;

  assert(c != NULL);
  assert(mean != NULL);
  assert(c->chip != NULL);

  numprobes = c->cdf->numprobes;

  for (i = 0; i < c->numchips; i++)
  {
    /* The rank is stored in the pm array */
    for (j = 0; j < numprobes; j++)
    {
      c->chip[i]->pm[j] = mean[(int)c->chip[i]->pm[j]];
    }
  }

  return;
}

int qnorm_compare(const void *p1, const void *p2)
{
  double x = ((dataitem *) p1)->data;
  double y = ((dataitem *) p2)->data;

  if (x < y)
    return -1;
  if (x > y)
    return 1;
  return 0;
}

/************************************************************
 **
 ** double *rank_order()
 **
 ** get ranks in the same manner as R does. Assume that *x is
 ** already sorted
 ** NB: Can this be in-place now?
 *************************************************************/

void rank_order(double *rank, dataitem * x, int n)
{
  int i, j, k;

  i = 0;

  while (i < n)
  {
    j = i;
    while ((j < n - 1) && (x[j].data == x[j + 1].data))
      j++;
    if (i != j)
    {
      for (k = i; k <= j; k++)
        rank[k] = (i + j + 2) / 2.0;
    }
    else
      rank[i] = i + 1;
    i = j + 1;
  }
  /*return rank; */
}
