
/**************************************************************************
 *
 * Filename: rma.c
 *
 * Purpose: 
 *
 * Creation date: 
 *
 * Author: Steven Eschrich
 *
 *
 * Update History
 * --------------
 * 4/14/05: Imported/repaired from old libaffy (AMH)
 * 4/19/05: AFFY_CELFILE now uses AFFY_CELL's (AMH)
 *
 **************************************************************************/

#include <affy_rma.h>

int rma_pb = 0;

AFFY_CHIPSET *rma(char **filelist, RMA_FLAGS *f)
{
  AFFY_CHIPSET *c;
  int i, j, k;
  char *chip_type;
  int max_chips;
  char **p;
  double *mean = NULL;
  int numprobes;
  struct hashtable *h = NULL;

  /* In case flags not used, create default entry */
  if (f == NULL)
    f = rma_get_defaults();

  /* Count up chips */
  for (p = filelist, max_chips = 0; *p != NULL; p++)
    max_chips++;

  /* Create structure */
  c = CALLOC(1, sizeof(AFFY_CHIPSET));
  c->chip = CALLOC(max_chips, sizeof(AFFY_CHIP *));

  /* Now load the cdf file (figure out chip type) */
  chip_type = affy_get_cdf_name_from_cel(filelist[0]);
  c->cdf = affy_load_cdf_file(chip_type, f->cdf_directory);
  c->numrows = c->cdf->numrows;
  c->numcols = c->cdf->numcols;

  numprobes = c->cdf->numprobes;

  if (f->intermediate_file != NULL) {
    /* Check for saved intermediate file */
    AFFY_PSLOAD psload;
    int np = 0;
    info(("Chiptype: %s\n", chip_type));
    if (init_probeset_loading(f->intermediate_file, c, &psload) == 0) {
      h = get_filenames(&psload);
      mean = CALLOC(c->cdf->numprobes, sizeof(double));
      bzero(mean, c->cdf->numprobes * sizeof(double));
      load_mean_from_disk(mean, &np, f->intermediate_file);
      assert(numprobes == np);
    }
    close_probeset_loading(&psload);
  }

  /* Load each chip */
  for (i = 0, j = 0; i < max_chips; i++)
  {

    double *pm;
    AFFY_CELL **data;
    AFFY_PROBE **probe_info = c->cdf->probe;

    if (h != NULL && hashtable_search(h,filelist[i]) != NULL) {
      /* CEL file found in the intermediate results */
      info(("Chip %d/%d (Skipped)... \n", i, max_chips));
      c->numchips++;
      c->chip[j] = CALLOC(1, sizeof(AFFY_CHIP));
      c->chip[j]->cel = NULL;
      c->chip[j]->filename = strdup(filelist[i]);
      c->chip[j]->cdf = c->cdf;
      j++;
      continue;
    }
    info(("Chip %d/%d ... \n", i, max_chips));
    /* Load chip */
    c->chip[j] = affy_load_chip(filelist[i]);
    if (c->chip[j] == NULL)
      continue;
    c->numchips++;

    /* Kludge: distribute cdf ptr */
    c->chip[j]->cdf = c->cdf;

    /* Load PM data, then free all data */
    pm = c->chip[j]->pm = CALLOC(c->cdf->numprobes, sizeof(double));
    data = c->chip[j]->cel->data;
    for (k = 0; k < c->cdf->numprobes; k++)
    {
      int x = probe_info[k]->pm.x;
      int y = probe_info[k]->pm.y;

      pm[k] = data[x][y].value;
#if 1
      if(isnan(pm[k])) {
	info(("Chip[%d] PM[%d] is NAN\n", j, k));
	exit(1);
      }
#endif
    }
    free(data[0]);
    free(data);
    c->chip[j]->cel->data = NULL;

    rma_pb = pb_create(5, "Processing chip");

    /* Background correct */
    if (f->use_background_correction)
    {
      rma_background_correct(c, j);
    }

    /* Normalize (partially) */
    if (f->use_normalization)
    {
      /* Option to use mean normalization */
      if (f->use_mean_normalization)
      {
	affy_mean_normalization(c, f->mean_normalization_target_value);
      }
      else
      {
	/* Default is quantile normalization */
	if (mean == NULL) {
	  mean = CALLOC(c->cdf->numprobes, sizeof(double));
	  bzero(mean, c->cdf->numprobes * sizeof(double));
	}
	quantile_normalization_chip(c, j, mean);
      }
    }

    if (f->intermediate_file != NULL) {
      /* DS: More efficient approach : save intermediate files to disk */
      /* save pm to disk */
      save_pm_to_disk(c, i, max_chips, f->intermediate_file);

      /* free memory */
      free(c->chip[i]->pm);
      c->chip[i]->pm = NULL;

      /* save mean to disk */
      save_mean_to_disk(mean, c->cdf->numprobes, f->intermediate_file);
    }

    pb_finish(rma_pb);
    /* Update counter of chips */
    j++;

  }

  /* With storage freed, create the smaller chipset */
  c->numchips = j;

  rma_pb = pb_create(100, "Processing chip");

  /* Redistribute quantile means */
  if (f->use_normalization && !f->use_mean_normalization)
  {
    for (i = 0; i < c->cdf->numprobes; i++)
    {
      mean[i] /= c->numchips;
    }
    c->mean = mean;
    if (f->intermediate_file == NULL) { /* Everything is in memory */
      quantile_normalization_chipset(c, mean);
      free(mean);
    }
  }

  /* Actually calculate the expression */
  rma_signal(c, f);

  if (f->intermediate_file != NULL && f->use_normalization &&
      !f->use_mean_normalization) {
    free(mean);
  }
  if (h != NULL) {
    hashtable_destroy(h,0);
  }
  pb_finish(rma_pb);

  /* Free up remaining space */
  for (i = 0; i < c->numchips; i++)
  {
    if (c->chip[i]->pm != NULL) {
      free(c->chip[i]->pm);
    }
    if (c->chip[i]->cel != NULL) {
      affy_free_cel_file(c->chip[i]->cel);
    }
  }
  return c;
}
