
/******************************************************************************
 **
 ** file: rma_background.c
 **
 ** Copyright (C) 2002 - 2003 B. M. Bolstad
 ** 
 ** Written by: B. M. Bolstad  <bolstad@stat.berkeley.edu>
 ** Implementation dates: 2002-2003
 */

#include <affy_rma.h>

extern int rma_pb;

/* Local functions */
#define PI 3.14159265358979323846

/* phi- compute the standard normal density. */
static INLINE double phi(double x)
{
  return 1 / sqrt(2 * PI) * exp(-0.5 * x * x);
}

/* Phi - Compute the standard normal distribution function */
static INLINE double Phi(double x)
{
  return affy_pnorm5(x, 0.0, 1.0, 1, 0);
}

double estimate_alpha();
double get_sd();

void estimate_bg_parameters();
void background_adjust();
double max_density();

/**
 * Background correct the values in pm 
 */
void rma_background_correct(AFFY_CHIPSET *c, int chipnum)
{
  double b;
  int j;
  double alpha, mu, sigma;
  double *pm = c->chip[chipnum]->pm;
  int n = c->cdf->numprobes;

  pb_update(rma_pb, 1, "RMA Background correction.");
  estimate_bg_parameters(pm, n, &alpha, &mu, &sigma);
  pb_update(rma_pb, 2, "");

  /* Adjust values. */
  b = mu + (alpha * sigma * sigma);
  for (j = 0; j < n; j++)
  {
    double a = pm[j] - b;

    pm[j] = a + sigma * phi(a / sigma) / Phi(a / sigma);
  }
  pb_update(rma_pb, 3, "");

  return;
}

void estimate_bg_parameters(double *pm, int n, double *alpha, double *mu,
                            double *sigma)
{
  int i = 0;
  double max;
  int numx = 0;
  double *x;

  x = CALLOC(n, sizeof(double));

  max = max_density(pm, n);

  /* Get all values less than max */
  numx = 0;
  for (i = 0; i < n; i++)
  {
    if (pm[i] < max)
      x[numx++] = pm[i];
  }
  /* Get max density and sd */
  max = max_density(x, numx);
  *sigma = get_sd(pm, n, max) * 0.85;

  /* Get all values greater than max */
  numx = 0;
  for (i = 0; i < n; i++)
  {
    if (pm[i] > max)
      x[numx++] = pm[i];
  }

  /* the 0.85 is to fix up constant in above */
  *alpha = estimate_alpha(x, numx, max);
  *mu = max;

  free(x);
}

/***************************************************************
 **
 ** double get_sd()
 ** 
 ** estimate the sigma parameter given vector x, value of maximum 
 ** of density of x, and length of x
 ***************************************************************/
double get_sd(double *x, int n, double max)
{
  double sigma;
  double tmpsum = 0.0;
  int numtop = 0;
  int i;

  for (i = 0; i < n; i++)
  {
    if (x[i] < max)
    {
      tmpsum += (x[i] - max) * (x[i] - max);
      numtop++;
    }
  }
  sigma = sqrt(tmpsum / (numtop - 1)) * sqrt(2.0) / 0.85;
  return sigma;
}

/***************************************************************
 **
 ** double  estimate_alpha(double *PM,double PMmax, int rows,int cols,int column)
 **
 ** estimate the alpha parameter given vector x value of maximum of density
 ** of x, dimensions of x matrix and column of interest using method proposed
 ** in affy2
 **
 **
 ***************************************************************/
double estimate_alpha(double *x, int length, double max)
{
  double alpha;
  int i;

  /* Adjust x by the max value */
  for (i = 0; i < length; i++)
    x[i] -= max;

  alpha = max_density(x, length);

  alpha = 1.0 / alpha;

  return alpha;
}
