
/**************************************************************************
 *
 * Filename: rma_signal.c
 *
 * Purpose: 
 *
 * Creation date: 
 *
 * Author: Steven Eschrich
 *
 *
 * Update History
 * --------------
 * 4/14/05: Imported/repaired from old libaffy (AMH)
 *
 **************************************************************************/

#include <affy_rma.h>

extern int rma_pb;

/**
 * This is the PM-only array implementation of rma median polish:
 * useful for low-memory applications.
 */
void rma_signal(AFFY_CHIPSET *c, RMA_FLAGS *f)
{
  int i, j;
  int ps;
  double *results;
  double *affinities;
  AFFY_CDFFILE *cdf;
  int numchips;
  double LOG2 = log(2.0);
  int affinity_size = 0;
  int progress, incr;
  AFFY_PSLOAD psload;

  /* Preconditions: The CHIPSET and the CDF exist */
  assert(c != NULL);
  assert(c->cdf != NULL);
  if (f == NULL)
    f = rma_get_defaults();

  /* Convenience variable */
  cdf = c->cdf;
  numchips = c->numchips;

  /* Allocate storage */
  for (i = 0; i < numchips; i++)
  {
    c->chip[i]->probe_set = CALLOC(cdf->numprobesets, sizeof(double));
    c->chip[i]->numprobesets = cdf->numprobesets;
  }
  results = CALLOC(numchips, sizeof(double));

  affinities = NULL;

  if (f->intermediate_file != NULL) {
    init_probeset_loading(f->intermediate_file, c, &psload);
  }
  /* Median polish each probeset. The results are stored in probeset */
  pb_update(rma_pb, 1, "Calculating expressions");
  progress = incr = cdf->numprobesets / 100;
  for (ps = 0; ps < cdf->numprobesets; ps++)
  {
    AFFY_PROBESET *p = &(cdf->probeset[ps]);
    int numprobes = p->numprobes;
    double **z;

    if (ps == progress)
    {
      pb_update(rma_pb, progress / incr, "");
      progress += incr;
    }

    /* Allocate storage for z matrix (for polishing) */
    z = create_matrix(numprobes, numchips);

    if (f->intermediate_file != NULL) {
      /* load matrix from disk */
      load_matrix_from_disk(&psload, ps, z);
    }
    else {
      /* Load logged data for this probe set */
      for (i = 0; i < numchips; i++)
      {
	/* Load probes from pm array or actual cel data */
	for (j = 0; j < numprobes; j++)
	{
	  z[j][i] = log(c->chip[i]->pm[p->probe[j].index]) / LOG2;
	}
      }
    }

    /* Even if affinities aren't stored, the median polish will record them */
    if (affinity_size < numprobes)
    {
      affinities = REALLOC(affinities, numprobes * sizeof(double));
      affinity_size = numprobes;
    }

    /* Do the median polish */
    median_polish(z, numprobes, numchips, results, affinities);

    free_matrix(z);

    /* Store results */
    for (i = 0; i < numchips; i++)
    {
      c->chip[i]->probe_set[ps] = results[i];
      if(isnan(results[i])) {
	info(("Chip[%d] Probeset[%d] is NAN\n", i, ps));
	exit(1);
      }
    }

    /* If desired, dump the probe affinities */
    if (f->dump_probe_affinities)
    {
      FILE *fp;
      char *af_name = f->affinities_filename;

      fp = FOPEN(af_name, "a");
      fprintf(fp, "%s", p->name);
      for (i = 0; i < numprobes; i++)
        fprintf(fp, "\t%g", affinities[i]);
      fprintf(fp, "\n");
      fclose(fp);
    }
  }
  if (f->intermediate_file != NULL) {
    close_probeset_loading(&psload);
  }
  /* Free up results storage */
  free(results);
  free(affinities);

  info(("done.\n"));

  return;
}
