
/**************************************************************************
 *
 * Filename: median.c
 *
 * Purpose: Statistical routines.
 *
 * Creation date: 
 *
 * Author: Steven Eschrich
 *
 *
 * Update History
 * --------------
 * 4/14/05: Imported/repaired from old libaffy (AMH)
 *
 **************************************************************************/

#include <utils.h>
#include <affy.h>

/**************************************************************************
 **
 ** double affy_median(double *x, int length)
 **
 ** double *x - vector
 ** int length - length of *x
 **
 ** returns the median of *x
 **
 *************************************************************************/

/*
  Copy the results over to a temporary array, then get the median.
  This preserves the initial ordering.
*/
double affy_median_save(double *x, int length)
{
  int i;
  double *buffer = MALLOC(length * sizeof(double));
  double result;

  for (i = 0; i < length; i++)
    buffer[i] = x[i];
  result = affy_median(buffer, length);
  free(buffer);

  return result;
}

/* 
   This function will sort x and find the median. Note this is 
   destructive, therefore you should call affy_median_save() if you 
   care about this.
*/
double affy_median(double *x, int length)
{
  int half;
  double med;

  qsort(x, length, sizeof(double), affy_median_sort);

  half = (length + 1) / 2;
  if (length % 2 == 1)
  {
    med = x[half - 1];
  }
  else
  {
    med = (x[half] + x[half - 1]) / 2.0;
  }

  return med;
}

/*****************************************************************************
 **
 ** void affy_get_row_median(double *z, double *rdelta, int rows, int cols)
 **
 ** double *z - matrix of dimension  rows*cols
 ** double *rdelta - on output will contain row medians (vector of length rows)
 ** int rows, cols - dimesion of matrix
 **
 ** get the row medians of a matrix 
 **
 *****************************************************************************/
void affy_get_row_median(double **z, double *rdelta, int rows, int columns)
{
  int i, j;
  double *buffer = MALLOC(columns * sizeof(double));

  for (i = 0; i < rows; i++)
  {
    for (j = 0; j < columns; j++)
      buffer[j] = z[i][j];

    rdelta[i] = affy_median(buffer, columns);
  }

  free(buffer);

  return;
}

/*****************************************************************************
 **
 ** void affy_get_col_median(double *z, double *cdelta, int rows, int cols)
 **
 ** double *z - matrix of dimension  rows*cols
 ** double *cdelta - on output will contain col medians (vector of length cols)
 ** int rows, cols - dimesion of matrix
 **
 ** get the col medians of a matrix 
 **
 *****************************************************************************/
void affy_get_column_median(double **z, double *cdelta, int rows,
                            int columns)
{

  int i, j;

  double *buffer = MALLOC(rows * sizeof(double));

  for (j = 0; j < columns; j++)
  {
    for (i = 0; i < rows; i++)
      buffer[i] = z[i][j];

    cdelta[j] = affy_median(buffer, rows);
  }

  free(buffer);
}

/* 
  Sorting function for qsort(). Given two double pointers, return the
  ordering.
*/
int affy_median_sort(const void *p1, const void *p2)
{
  double n1 = *((double *)p1);
  double n2 = *((double *)p2);

  if (n1 < n2)
    return -1;
  if (n1 > n2)
    return 1;
  return 0;
}
