
/**************************************************************************
 *
 * Filename: pixels_from_cell.c
 *
 * Purpose: Obtain an AFFY_PIXREGION for a given cell.
 *
 * Creation date: 21 Apr 2005 
 *
 * Author: Andrew M. Hoerter
 *
 *
 * Update History
 * --------------
 * 4/21/05: Creation (AMH)
 *
 **************************************************************************/

#include <affy.h>

/* 
 * affy_pixels_from_cell(): Return the AFFY_PIXREGION corresponding to a 
 *                          cell.
 *
 * Inputs: cp points to an initialized/loaded AFFY_CHIP, x/y specify the
 *         cell coordinates. 
 *        
 *
 * Outputs: Pointer to an AFFY_PIXREGION or NULL on error.
 *
 * Side effects: The AFFY_CELL structure is updated to cache the result.
 *
 * TODO: Boundary conditions.
 *
 *
 */
AFFY_PIXREGION *affy_pixels_from_cell(AFFY_CHIP *cp, int x, int y)
{
	AFFY_POINT px, px_r,px_l;
	AFFY_PIXREGION *px_result=calloc(1,sizeof(AFFY_PIXREGION));
	int i;

  assert(cp != NULL);
  assert(cp->dat != NULL);

	/* Carefully checked if already cached */
	if ( cp->cel != NULL && cp->cel->data != NULL & cp->cel->data[x][y].pixels != NULL)
		return cp->cel->data[x][y].pixels;
			
	/* First get the starting position in image, and the right/lower boundaries */
	px=affy_cell_to_pixel(cp,x,y);
	px_r=affy_cell_to_pixel(cp,x+1,y);
	px_l=affy_cell_to_pixel(cp,x,y+1);

	px_result->numcols=px_r.x-px.x;
	px_result->numrows=px_l.y-px.y;
	px_result->data=calloc(px_result->numrows,sizeof(unsigned int *));
	for (i=0; i < px_result->numrows; i++) {
		px_result->data[i]=&(cp->dat->pixels.data[px.y+i][px.x]);
  }		
  
	if ( cp->cel != NULL )
	  cp->cel->data[x][y].pixels=px_result;

	return px_result;
}

/*
 * Convert a cell location (a cell is a collection of pixels) to image
 * coordinates. This requires rotation of the image which is done in
 * real time via a bilinear algorithm
 * 
 * Detecting and Correting Misalignment in Affymetrix Data
 * Keith A. Baggerly
 * Technical Report
 * MD Anderson Cancer Center
 */
AFFY_POINT affy_cell_to_pixel(AFFY_CHIP *cp,int x, int y)
{
	int ax,bx,cx,dx,ay,by,cy,dy;
	double rows,cols;
	double xn,yn,xp,yp,newx,newy;
	AFFY_POINT newp;
	
	ax=cp->dat->grid_ul.x;
	ay=cp->dat->grid_ul.y;
	bx=cp->dat->grid_ur.x;
	by=cp->dat->grid_ur.y;
	cx=cp->dat->grid_ll.x;
	cy=cp->dat->grid_ll.y;
	dx=cp->dat->grid_lr.x;
	dy=cp->dat->grid_lr.y;
	/* Corners are off, so fix */
	bx++;dx++;cy++;dy++;
	/* Get chip-specific rows/cols */
	rows=cp->cdf->numrows;
	cols=cp->cdf->numcols;
	
	/* Calculate x/y coordinates for current position, and one below and to right */
	xn=ax*((cols-x)/cols)*((rows-y)/rows) +
			bx*(x/cols)*((rows-y)/rows) +
			cx*((cols-x)/cols)*(y/rows) +
			dx*(x/cols)*(y/rows);
			
	yn=ay*((cols-x)/cols)*((rows-y)/rows) +
			by*(x/cols)*((rows-y)/rows) +
			cy*((cols-x)/cols)*(y/rows) +
			dy*(x/cols)*(y/rows);
	y++;
  xp=ax*((cols-x)/cols)*((rows-y)/rows) + 
        bx*(x/cols)*((rows-y)/rows) +
        cx*((cols-x)/cols)*(y/rows) +
        dx*(x/cols)*(y/rows);
  y--;
  x++;
  yp=ay*((cols-x)/cols)*((rows-y)/rows) + 
        by*(x/cols)*((rows-y)/rows) +
        cy*((cols-x)/cols)*(y/rows) +
        dy*(x/cols)*(y/rows);
  x--;

	/* Take the average of the current position and next one */
  newx=(xn+xp)/2.0;
  newy=(yn+yp)/2.0;

	/* Assign new values to a point, to return */
	newp.x=(int)(newx+0.5);
	newp.y=(int)(newy+0.5);

	return newp;
}
