
/**************************************************************************
 *
 * Filename: quantile_normalizations.c
 *
 * Purpose: Statistical routines.
 *
 * Creation date: 
 *
 * Author: Steven Eschrich
 *
 *
 * Update History
 * --------------
 * 4/8/05: Imported/repaired from old libaffy (AMH)
 * 4/19/05: AFFY_CELFILE now uses AFFY_CELL's (AMH)
 *
 **************************************************************************/

#include <affy.h>

/**********************************************************
 **
 ** file: qnorm.c
 **
 ** aim: A c implementation of the quantile normalization method 
 **
 ** Copyright (C) 2002-2003    Ben Bolstad
 **
 ** written by: B. M. Bolstad  <bolstad@stat.berkeley.edu>
 **
 ** written: Feb 2, 2002
 ** last modified: Apr 19, 2002
 ** 
 ** This c code implements the quantile normalization method
 ** for normalizing high density oligonucleotide data as discussed
 ** in
 **
 ** Bolstad, B. M., Irizarry R. A., Astrand, M, and Speed, T. P. (2003)(2003) 
 ** A Comparison of Normalization Methods for High 
 ** Density Oligonucleotide Array Data Based on Bias and Variance.
 ** Bioinformatics 19,2,pp 185-193
 **
 ** History
 ** Feb 2, 2002 - Intial c code version from original R code
 ** Apr 19, 2002 - Update to deal more correctly with ties (equal rank)
 ** Jan 2, 2003 - Documentation/Commenting updates reformating
 ** Feb 17, 2003 - add in a free(datvec) to qnorm(). clean up freeing of dimat
 ** Feb 25, 2003 - try to reduce or eliminate compiler warnings (with gcc -Wall)
 ** Feb 28, 2003 - update reference to normalization paper in comments
 ** Mar 25, 2003 - ability to use median, rather than mean in so called "robust" method
 **
 ***********************************************************/

/*
 * This normalizes both PM and MM values, which is different from the RMA
 * version, which only normalizes the PM values.
 */

void affy_quantile_normalization(AFFY_CHIPSET *d)
{
  int i, j, x, y;
  double sum;
  double *rank, *mean;
  AFFY_CELFILE *cf;
  int number_of_probes;
  double ***all_vals;

  info("Quantile normalization...");

  number_of_probes = 2 * d->cdf->number_of_probes;

  /* Allocate storage */
  mean = CALLOC(number_of_probes, sizeof(double));
  all_vals = CALLOC(d->number_of_chips, sizeof(double **));

  /* We start by creating a matrix of double pointers to data */
  for (i = 0; i < d->number_of_chips; i++)
  {
    cf = d->chip[i]->cel;
    j = 0;

    /* Allocate storage for this chip */
    all_vals[i] = CALLOC(number_of_probes, sizeof(double *));

    /* Load in both the pm and mm for each probe */
    for (j = 0; j < number_of_probes / 2; j++)
    {
      x = d->cdf->probes[j]->pm.x;
      y = d->cdf->probes[j]->pm.y;
      all_vals[i][2 * j] = &(cf->data[x][y].value);
      x = d->cdf->probes[j]->mm.x;
      y = d->cdf->probes[j]->mm.y;
      all_vals[i][2 * j + 1] = &(cf->data[x][y].value);
    }

    qsort(all_vals[i], number_of_probes, sizeof(double *), qnorm_compare);
  }

  /* Step two: calculate mean value at a given rank */
  for (j = 0; j < number_of_probes; j++)
  {
    sum = 0;
    /* Calculate mean */
    for (i = 0; i < d->number_of_chips; i++)
      sum += *(all_vals[i][j]) / (double)d->number_of_chips;

    mean[j] = sum;
  }

  rank = CALLOC(number_of_probes, sizeof(double));

  /* Step 3: Redistribute mean value to all chips */
  for (i = 0; i < d->number_of_chips; i++)
  {
    /* Rank order the intensities on this chip */
    rank_order(rank, all_vals[i], number_of_probes);
    for (j = 0; j < number_of_probes; j++)
    {
      *(all_vals[i][j]) = mean[(int)floor(rank[j]) - 1];
    }
    free(all_vals[i]);
  }
  free(rank);

  free(mean);
  free(all_vals);

  info("done.\n");

  return;
}

int affy_qnorm_compare(const void *p1, const void *p2)
{
  double x = **((double **)p1);
  double y = **((double **)p2);

  if (x < y)
    return -1;
  if (x > y)
    return 1;
  return 0;
}

/************************************************************
 **
 ** double *affy_rank_order()
 **
 ** get ranks in the same manner as R does. Assume that *x is
 ** already sorted
 **
 *************************************************************/

void affy_rank_order(double *rank, double **x, int n)
{
  int i, j, k;

  i = 0;

  while (i < n)
  {
    j = i;
    while ((j < n - 1) && (*(x[j]) == *(x[j + 1])))
      j++;
    if (i != j)
    {
      for (k = i; k <= j; k++)
        rank[k] = (i + j + 2) / 2.0;
    }
    else
      rank[i] = i + 1;
    i = j + 1;
  }
  /*return rank; */
}
