/**************************************************************************
 *
 * Filename: buffered_reader.c
 *
 * Purpose: Create a low level buffered reader
 *
 * Creation date: 5/14/07
 *
 * Author: Debashis Sahoo
 *
 *
 * Update History
 * --------------
 * 5/14/07: Created
 *
 * $Id:$
 *
 **************************************************************************/

#include "buffered_reader.h"
#include "utils.h"

buffered_reader * init_buffered_reader(FILE *fp, int size) {
  buffered_reader * res = NULL; 
  res = CALLOC(1, sizeof(buffered_reader));
  bzero(res, sizeof(buffered_reader));
  res->fp = fp;
  res->bufsize = size;
  res->buffer = CALLOC(size, sizeof(char));
  return res;
}

void close_buffered_reader(buffered_reader *res) {
  if (res) {
    free(res->buffer);
    free(res);
  }
}

int read_byte(buffered_reader * b) {
  int res;
  if (b->startpos >= b->endpos) {
    if (fill_buffer(b) < 0) {
      return -1;
    }
  }
  if (b->endpos == 0) {
    return -1;
  }
  else {
    res = b->buffer[b->startpos] & 0xff;
    b->startpos++;
    return res;
  }
}

int fill_buffer(buffered_reader * b) {
  int n = 0;
  if(b->curpos != ftell(b->fp)) {
    if (fseek(b->fp, b->curpos, SEEK_SET) < 0) {
      return -1;
    }
  }
  n = fread(b->buffer, 1, b->bufsize, b->fp);
  if (n >= 0) {
    b->curpos += n;
    b->endpos = n;
    b->startpos = 0;
  }
  if (b->startpos >= b->endpos) {
    return -1;
  }
  return n;
}

void invalidate(buffered_reader * b) {
  b->endpos = 0;
  b->startpos = 0;
  b->curpos = ftell(b->fp);
}

int read_buffered(buffered_reader * b, char *buf, int offset, int len) {
  int leftover = b->endpos - b->startpos;
  int i, c;
  if (len <= leftover) {
    memcpy(buf+offset, b->buffer+ b->startpos, len);
    b->startpos += len;
    return len;
  }
  for(i = 0; i < len; i++) {
    c = read_byte(b);
    if(c != -1)
      buf[offset+i] = (char)c;
    else {
      return i;
    }
  }
  return len;
}

long get_filepointer(buffered_reader * b) {
  long l = b->curpos;
  return (l - b->endpos + b->startpos);
}

int seek_buffered(buffered_reader * b, long pos) {
  int n;
  n = (int)(b->curpos - pos);
  if (n >= 0 && n <= b->endpos) {
    b->startpos = b->endpos - n;
  }
  else {
    if (fseek(b->fp, pos, SEEK_SET) < 0) {
      return -1;
    }
    invalidate(b);
  }
  return 0;
}

int readline_buffered(buffered_reader * b, char * buf, int size) {
  int lineend = -1, i = 0, len = 0;
  if (b->endpos <= b->startpos) {
    if (fill_buffer(b) < 0) {
      return -1;
    }
  }
  for (i= b->startpos; i < b->endpos; i++) {
    if (b->buffer[i] == '\n') {
      lineend = i;
      break;
    }
  }
  if (lineend < 0) { /* New line is not in buffer */
    for (i = 0; i < size; i++) {
      buf[i] = (char) read_byte(b);
      if (buf[i] == '\n') {
	return i+1;
      }
    }
    return size;
  }
  for (i = 0; i < size; i++) {
    buf[i] = b->buffer[b->startpos + i];
    if (buf[i] == '\n') {
      i++;
      break;
    }
  }
  len = i;
  b->startpos = b->startpos + i;
  return len;
}

